import numpy as np
import sys
import time
import matplotlib.pyplot as plt
import pickle
import glob
import torch
import torch.nn.init as init
import networkx as nx
from .circuit_utils import Circuit


class RBM(object):
    """Class for generic Restricted Boltzmann Machine (RBM)

    Parameters
    ----------
    model_name : str
        Name of the model
    n_vis : int
        Number of visible units
    n_hin : int
        Number of hidden units
    k : int
        Number of Gibbs sampling steps
    minMax_W : tuple
        Tuple containing the minimum and maximum values for the weights
    energy_type : str
        Type of energy_type function to use. Options are 'hopfield' and 'linear_circuit_i' with i=1,2,3
    optimizer : str
        Type of optimizer to use. Options are 'Adam' and 'SGD'
    regularization : bool
        Whether to use regularization or not
    l1_factor : float
        L1 regularization factor
    l2_factor : float
        L2 regularization factor

    Attributes
    ----------
    name : str
        Name of the model
    W : array-like, shape (n_hin, n_vis)
        Weight matrix
    v_bias : array-like, shape (n_vis,)
        Visible bias vector
    h_bias : array-like, shape (n_hin,)
        Hidden bias vector
    k : int
        Number of Gibbs sampling steps
    min_W : float
        Minimum value for the weights
    max_W : float
        Maximum value for the weights
    energy_type : str
        Type of energy_type function to use. Options are 'hopfield' and 'linear_circuit_i' with i=4,5
    n_hidden : int
        Number of hidden units
    n_visible : int
        Number of visible units
    epoch : int
        Current epoch
    errors_free_energy : list
        List containing the free energy difference between data and model
    errors_loss : list
        List containing the loss between data and model
    regularization : bool
        Whether to use regularization or not
    l1 : float
        L1 regularization factor
    l2 : float
        L2 regularization factor
    optimizer : str
        Type of optimizer to use. Options are 'Adam' and 'SGD'
    lr : float
        Learning rate
    m_dW : float
        Adam's momentum for the weights
    m_dv : float
        Adam's momentum for the visible bias
    m_dh : float
        Adam's momentum for the hidden bias
    v_dW : float
        Adam's velocity for the weights
    v_dv : float
        Adam's velocity for the visible bias
    v_dh : float
        Adam's velocity for the hidden bias
    beta1 : float
        Adam's beta1 parameter
    beta2 : float
        Adam's beta2 parameter
    epsilon : float
        Adam's epsilon parameter
    """

    def __init__(
        self,
        model_name,
        n_vis=784,
        n_hin=50,
        k=1,
        lr=0.01,
        max_epochs=200000,
        minMax_W=(-100, 100),
        energy_type='hopfield',
        optimizer='Adam',
        regularization=False,
        l1_factor=1e-7,
        l2_factor=1e-7,
        ground_v=1,
        ground_h=1,
        batch_size=64,
        train_algo='CD',
        centering=False,
        average_data=None,
        sampling_model_beta=1,
        mytype=torch.double,
    ):
        self.name = model_name
        print('*** Initializing {}'.format(self.name))
        self.device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu')
        # Set default dtype
        self.mytype = mytype
        torch.set_default_dtype(self.mytype)
        print('The model is working on the following device: {}'.format(
            self.device))
        self.k = k
        self.max_epochs = max_epochs
        self.lr = lr
        self.min_W, self.max_W = torch.Tensor(minMax_W).to(self.device)
        self.energy_type = energy_type
        self.n_hidden = n_hin
        self.n_visible = n_vis
        self.model_beta = sampling_model_beta
        # Quantities to store
        self.epoch = 0
        self.regularization = regularization
        if self.regularization == 'l1':
            self.l1 = l1_factor
        elif self.regularization == 'l2':
            self.l2 = l2_factor
        self.optimizer = optimizer
        # Initialize weights
        self.W = torch.randn(
            (
                n_hin,
                n_vis,
            ),
            dtype=self.mytype,
            device=self.device,
        )
        self.v_bias = torch.zeros((n_vis, ),
                                  dtype=self.mytype,
                                  device=self.device)
        self.h_bias = torch.randn((n_hin, ),
                                  dtype=self.mytype,
                                  device=self.device)
        # Make weights contiguous
        self.W = self.W.contiguous()
        self.v_bias = self.v_bias.contiguous()
        self.h_bias = self.h_bias.contiguous()
        ## Initialize with normal distribution
        init.xavier_normal_(self.W)
        init.normal_(self.v_bias)
        init.normal_(self.h_bias)
        if average_data is not None:
            # Initialize the visible bias from data frequency
            self.v_bias = (torch.log(average_data / (1 - average_data) +
                                     1e-5).to(self.device).to(self.mytype))
        if self.optimizer == 'Adam':
            # Adam's momenta
            self.m_dW = 0 * self.W
            self.m_dv = 0 * self.v_bias
            self.m_dh = 0 * self.h_bias
            self.v_dW = 0 * self.W
            self.v_dv = 0 * self.v_bias
            self.v_dh = 0 * self.h_bias
            # Adam's parameters
            self.beta1 = 0.9
            self.beta2 = 0.999
            self.epsilon = 1e-8
        if self.energy_type != 'hopfield':
            self.ground_v = ground_v
            self.ground_h = ground_h
            self._update_circuit_variables()
            if self.energy_type == 'linear_circuit_5':
                # Create the circuit
                self.circuit = nx.Graph()

                # add two visible nodes for each original visible node and connect them to the hidden nodes and the bias node
                # One loop per each set of nodes to ensure the correct order in networkx
                for i in range(self.n_visible):
                    self.circuit.add_node(i)  # positive visible node
                for i in range(self.n_visible):
                    self.circuit.add_node(n_vis + i)  # negative visible node
                for j in range(self.n_hidden):
                    self.circuit.add_node(2 * n_vis +
                                          j)  # positive hidden node
                for j in range(self.n_hidden):
                    self.circuit.add_node(2 * n_vis + n_hin +
                                          j)  # negative hidden node

                # add bias nodes
                self.circuit.add_node(2 * n_vis + 2 * n_hin)
                self.circuit.add_node(2 * n_vis + 2 * n_hin + 1)

                # add edges
                for i in range(self.n_visible):
                    for j in range(self.n_hidden):
                        self.circuit.add_edge(i, 2 * n_vis +
                                              j)  # , conductance=k_plus[j][i])
                        self.circuit.add_edge(
                            i, 2 * n_vis + n_hin +
                            j)  # , conductance=k_minus[j][i])
                        self.circuit.add_edge(
                            n_vis + i,
                            2 * n_vis + j)  # , conductance=k_minus[j][i])
                        self.circuit.add_edge(n_vis + i, 2 * n_vis + n_hin +
                                              j)  # , conductance=k_plus[j][i])
                    self.circuit.add_edge(
                        i, 2 * n_vis + 2 * n_hin)  # , conductance=q_plus[i])
                    self.circuit.add_edge(
                        n_vis + i,
                        2 * n_vis + 2 * n_hin)  # , conductance=q_minus[i])

                for j in range(self.n_hidden
                               ):  # connect the bias node to all hidden nodes
                    self.circuit.add_edge(2 * n_vis + j,
                                          2 * n_vis + 2 * n_hin +
                                          1)  # , conductance=c_plus[j])
                    self.circuit.add_edge(2 * n_vis + n_hin + j,
                                          2 * n_vis + 2 * n_hin +
                                          1)  # , conductance=c_minus[j])

                # assign the positions into three columns: bias, visible, hidden. All the columns are aligned horizontally.
                shift_vis = (self.n_visible - 1) / 2
                shift_hid = (self.n_hidden - 1) / 2

                spacing = 0.3
                adjusted_positions_vis = [(i + j * spacing)
                                          for i in range(self.n_visible)
                                          for j in range(2)]
                adjusted_positions_hid = [(i + j * spacing)
                                          for i in range(self.n_hidden)
                                          for j in range(2)]

                # normalize the positions to span 0 to (self.n_visible - 1) for visible and (self.n_hidden - 1) for hidden
                min_pos_vis = min(adjusted_positions_vis)
                max_pos_vis = max(adjusted_positions_vis)
                normalized_positions_vis = [
                    (pos - min_pos_vis) * (self.n_visible - 1) /
                    (max_pos_vis - min_pos_vis)
                    for pos in adjusted_positions_vis
                ]

                min_pos_hid = min(adjusted_positions_hid)
                max_pos_hid = max(adjusted_positions_hid)
                normalized_positions_hid = [
                    (pos - min_pos_hid) * (self.n_hidden - 1) /
                    (max_pos_hid - min_pos_hid)
                    for pos in adjusted_positions_hid
                ]

                pos = {
                    2 * n_vis + 2 * n_hin: (0, 0),
                    2 * n_vis + 2 * n_hin + 1: (3, 0)
                }
                pos.update({
                    i: (1, normalized_positions_vis[2 * i] - shift_vis)
                    for i in range(n_vis)
                })
                pos.update({
                    n_vis + i:
                    (1, normalized_positions_vis[2 * i + 1] - shift_vis)
                    for i in range(n_vis)
                })
                pos.update({
                    2 * n_vis + j:
                    (2, normalized_positions_hid[2 * j] - shift_hid)
                    for j in range(n_hin)
                })
                pos.update({
                    2 * n_vis + n_hin + j: (
                        2,
                        normalized_positions_hid[2 * j + 1] - shift_hid,
                    )
                    for j in range(n_hin)
                })

                # Set the positions of the nodes as attributes
                nx.set_node_attributes(self.circuit, pos, 'pos')

                circuit = Circuit(self.circuit)
                self.circuit = circuit

                # Store the indices of the visible nodes, hidden nodes, and bias nodes
                self.v_indices_plus = np.arange(self.n_visible)
                self.v_indices_minus = np.arange(self.n_visible,
                                                 2 * self.n_visible)
                self.h_indices_plus = np.arange(
                    2 * self.n_visible, 2 * self.n_visible + self.n_hidden)
                self.h_indices_minus = np.arange(
                    2 * self.n_visible + self.n_hidden,
                    2 * self.n_visible + 2 * self.n_hidden,
                )
                self.g_indices = np.arange(
                    2 * self.n_visible + 2 * self.n_hidden,
                    2 * self.n_visible + 2 * self.n_hidden + 2,
                )

                # Store the indices of the edges mapping from the ordered interactions to the edge list of networkx
                AA, BB = np.meshgrid(self.v_indices_plus, self.h_indices_plus)
                C1 = np.column_stack((AA.ravel(), BB.ravel()))

                AA, BB = np.meshgrid(self.v_indices_minus,
                                     self.h_indices_minus)
                C2 = np.column_stack((AA.ravel(), BB.ravel()))

                AA, BB = np.meshgrid(self.v_indices_plus, self.h_indices_minus)
                C3 = np.column_stack((AA.ravel(), BB.ravel()))

                AA, BB = np.meshgrid(self.v_indices_minus, self.h_indices_plus)
                C4 = np.column_stack((AA.ravel(), BB.ravel()))

                AA, BB = np.meshgrid(self.v_indices_plus, self.g_indices[0])
                C5 = np.column_stack((AA.ravel(), BB.ravel()))

                AA, BB = np.meshgrid(self.v_indices_minus, self.g_indices[0])
                C6 = np.column_stack((AA.ravel(), BB.ravel()))

                AA, BB = np.meshgrid(self.h_indices_plus, self.g_indices[1])
                C7 = np.column_stack((AA.ravel(), BB.ravel()))

                AA, BB = np.meshgrid(self.h_indices_minus, self.g_indices[1])
                C8 = np.column_stack((AA.ravel(), BB.ravel()))

                C = np.concatenate((C1, C2, C3, C4, C5, C6, C7, C8), axis=0)
                edgelist = np.array(self.circuit.graph.edges)

                edge_indexmap = []
                # For each row in edgelist, find the row in C that matches
                for row in edgelist:
                    # Use broadcasting to find the row in A
                    # np.all with axis 1 makes sure all elements in a row are equal
                    index = np.where((C == row).all(axis=1))[0][0]
                    edge_indexmap.append(index)

                self.edge_indexmap = np.array(edge_indexmap)

                self.Q_visible_clamped = self.circuit.constraint_matrix(
                    np.concatenate((self.v_indices_plus, self.v_indices_minus,
                                    self.g_indices)))
                self.Q_hidden_clamped = self.circuit.constraint_matrix(
                    np.concatenate((self.h_indices_plus, self.h_indices_minus,
                                    self.g_indices)))

                self.min_conductance = 0

        self.train_algo = train_algo
        self.batch_size = batch_size
        if self.train_algo == 'PCD':
            # Initialize the persistent chains
            self.persistent_chains = (torch.where(
                torch.rand(self.batch_size, self.n_visible) > 0.5, 1.0,
                0.0).to(self.device).to(self.mytype))
        self.centering = centering
        if self.centering:
            if average_data.shape[0] != n_vis:
                print(
                    'Error: you need to provide the average of the data to center the gradient'
                )
                sys.exit()
            # Initialize the offsets for the gradient centering
            self.ov = average_data.to(self.device)
            self.oh = self.h_bias * 0 + 0.5
            self.batch_ov = self.v_bias * 0
            self.batch_oh = self.h_bias * 0
            # And the sliding factors
            self.slv = 0.01
            self.slh = 0.01
        else:
            self.ov = 0
            self.oh = 0
        # Epochs at which to store the model
        num_points = 50
        self.t_to_save = sorted(
            list(
                set(
                    np.round(
                        np.logspace(np.log10(1), np.log10(self.max_epochs),
                                    num_points)).astype(int).tolist())))

        #### Physical performance ####

        self.power_f = 0
        self.power_b = 0
        self.energy = 0
        self.W_t = self.W.t()
        self.relax_t_f, self.relax_t_b = self.relaxation_times()

    def pretrain(self, pretrained_model):
        # Check if you have model load points
        filename_list = glob.glob(
            'model_states/{}_t*.pkl'.format(pretrained_model))
        if len(filename_list) > 0:
            all_loadpoints = sorted([
                int(x.split('_t')[-1].split('.pkl')[0]) for x in filename_list
            ])
            last_epoch = all_loadpoints[-1]
            print('** Using as pretraining model {} at epoch {}'.format(
                pretrained_model, last_epoch))
            with open(
                    'model_states/{}_t{}.pkl'.format(pretrained_model,
                                                     last_epoch),
                    'rb') as file:
                temp_model = pickle.load(file)
                # *** Import pretrained parameters
                self.W = temp_model.W.to(self.mytype)
                self.h_bias = temp_model.h_bias.to(self.mytype)
                self.v_bias = temp_model.v_bias.to(self.mytype)
        else:
            print('** No load points for {}'.format(pretrained_model))

    def v_to_h(self, v, beta=None):
        if beta is None:
            beta = self.model_beta
        if self.energy_type == 'linear_circuit_5':
            return self.Bernoulli_v_to_h(v, beta)
        else:
            if beta > 100:
                # I assume we are at T=0
                return self.Deterministic_v_to_h(v, beta)
            else:
                return self.Bernoulli_v_to_h(v, beta)

    def h_to_v(self, h, beta=None):
        if beta is None:
            beta = self.model_beta
        if self.energy_type == 'linear_circuit_5':
            return self.Bernoulli_h_to_v(h, beta)
        else:
            if beta > 100:
                # I assume we are at T=0
                return self.Deterministic_h_to_v(h, beta)
            else:
                return self.Bernoulli_h_to_v(h, beta)

    def Deterministic_v_to_h(self, v, beta):
        h = (self.delta_eh(v) > 0).to(v.dtype)
        return h, h

    def Deterministic_h_to_v(self, h, beta):
        v = (self.delta_ev(h) > 0).to(h.dtype)
        return v, v

    def Bernoulli_v_to_h(self, v, beta):
        p_h = self._prob_h_given_v(v, beta)
        sample_h = torch.bernoulli(p_h)
        return p_h, sample_h

    def Bernoulli_h_to_v(self, h, beta):
        p_v = self._prob_v_given_h(h, beta)
        sample_v = torch.bernoulli(p_v)
        return p_v, sample_v

    def _free_energy_hopfield(self, v, beta=None):
        if beta is None:
            beta = self.model_beta
        vbias_term = torch.mv(v, self.v_bias) * beta
        wx_b = torch.mm(v, self.W.t()) + self.h_bias
        hidden_term = torch.sum(torch.log(1 + torch.exp(wx_b * beta)), axis=1)
        return -hidden_term - vbias_term

    def _energy_hopfield(self, v, h):
        energy = (-(torch.mm(v, self.W.t()) * h).sum(1) -
                  torch.mv(v, self.v_bias) - torch.mv(h, self.h_bias))
        return energy

    def _energy_linear_circuit_4(self, v, h):
        # ****** TO BE DONE! ******
        print('Warning: not implemented')
        energy = (-(torch.mm(v, self.W.t()) * h).sum(1) -
                  torch.mv(v, self.v_bias) - torch.mv(h, self.h_bias))
        return energy

    def forward(self, v, k, beta=None):
        if beta is None:
            beta = self.model_beta
        pre_h1, h1 = self.v_to_h(v, beta)
        h_ = h1
        for _ in range(k):
            pre_v_, v_ = self.h_to_v(h_, beta)
            pre_h_, h_ = self.v_to_h(v_, beta)
        return v_

    def train(
        self,
        train_data,
        test_data=[],
        print_error=False,
        print_test_error=False,
    ):
        """Train the model using the given data and parameters."""
        while self.epoch < self.max_epochs:
            self.tstep_keepup()
            # permutation = torch.randperm(train_data.size(0))
            # train_data = train_data[permutation]
            for _, v_data in enumerate(train_data):
                # n_batches = int(train_data.shape[0]/self.batch_size)
                # for bi in range(n_batches):
                #    v_data = train_data[bi*self.batch_size:(bi+1)*self.batch_size]
                start_time = time.time()
                if self.energy_type == 'linear_circuit_5':
                    # self._weights_to_conductances()
                    # set the conductances in the same order as they are internally stored
                    # conductances = [self.circuit.graph.edges[e]['conductance'] for e in self.circuit.graph.edges]
                    self.circuit.setConductances(
                        self._weights_to_conductances())
                if self.train_algo == 'PCD':
                    # Update the chain after every batch
                    self.persistent_chains = self.forward(
                        self.persistent_chains, self.k)
                    v_model = self.persistent_chains
                elif self.train_algo == 'RDM':
                    # This algo uses random samples
                    # But since we want to train with the same exact protocol that we will use for generation,
                    ## we random sample from the hidden and not the visible
                    # h_rnd = torch.randint(high=2, size=(self.batch_size, self.n_hidden), device=self.device, dtype=self.mytype)
                    # _, v_model = self.h_to_v(h_rnd)
                    v_model = torch.randint(
                        high=2,
                        size=(self.batch_size, self.n_visible),
                        device=self.device,
                        dtype=self.mytype,
                    )
                    # v_model = torch.bernoulli(0.5*torch.ones(size=(self.batch_size, self.n_visible), device=self.device, dtype=self.mytype))
                    #### This version has problems
                    ###random_tensor = torch.rand(self.batch_size, self.n_visible, device=self.device, dtype=self.mytype)
                    ###v_model = self.forward(torch.where(random_tensor > 0.5, 1.0, 0.0), self.k)
                    v_model = self.forward(v_model, self.k)
                elif self.train_algo == 'CD':
                    v_model = self.forward(v_data, self.k)

                # Apply model
                h_data = self.v_to_h(v_data)[0]
                h_model = self.v_to_h(v_model)[0]

                # compute power and update energy
                p_f = self.power_forward(v_data, g_v=0.5, g_h=0.5)
                p_b = self.power_backward(h_model, g_v=0.5, g_h=0.5)
                self.power_f += p_f.mean()
                self.power_b += p_b.mean()
                self.energy += p_f.sum() + p_b.sum()

                # Apply centering
                if self.centering:
                    self.batch_ov = v_data.mean(0)
                    self.batch_oh = h_data.mean(0)
                    # update with sliding
                    self.ov = (1 -
                               self.slv) * self.ov + self.slv * self.batch_ov
                    self.oh = (1 -
                               self.slh) * self.oh + self.slh * self.batch_oh

                # Compute gradients
                partial_gradients_data = self.derivatives(v_data, h_data)
                partial_gradients_model = self.derivatives(v_model, h_model)

                # Average gradient over batch
                partial_gradients_data = self.average_gradients(
                    partial_gradients_data)
                partial_gradients_model = self.average_gradients(
                    partial_gradients_model)

                # Update weights and biases
                if self.optimizer == 'Adam':
                    self.Adam_update(
                        self.epoch + 1,
                        partial_gradients_data,
                        partial_gradients_model,
                    )
                elif self.optimizer == 'SGD':
                    self.SGD_update(
                        partial_gradients_data,
                        partial_gradients_model,
                    )

                self.tstep_keepup()

                self.epoch += 1

                # Store the model state
                if self.epoch in self.t_to_save:
                    with open(
                            'model_states/{}_t{}.pkl'.format(
                                self.name, self.epoch), 'wb') as file:
                        pickle.dump(self, file)

                if self.epoch % 100 == 0:
                    t = time.time() - start_time
                    if print_error:
                        v_model = self.forward(v_data, 1)
                        rec_error_train = ((v_model -
                                            v_data)**2).mean(1).mean(0)
                        if not print_test_error:
                            print('Epoch: %d , train-err %.5g , time: %f' %
                                  (self.epoch, rec_error_train, t))
                        else:
                            t_model = self.forward(test_data, 1)
                            rec_error_test = (((t_model -
                                                test_data)**2).mean(1).mean(0))
                            print(
                                'Epoch: %d , Test-err %.5g , train-err %.5g , time: %f'
                                % (self.epoch, rec_error_test, rec_error_train,
                                   t))
                    else:
                        print('Epoch: %d , time: %f' % (self.epoch, t))

        print('*** Training finished')

    def tstep_keepup(self):
        if self.energy_type != 'hopfield':
            self.clip_weights()
            self.clip_bias()
            self._update_circuit_variables()
        else:
            self.W_t = self.W.t()
        self.relax_t_f, self.relax_t_b = self.relaxation_times()

    def average_gradients(self, gradients):
        average_gradients = []
        for g in gradients:
            average_gradients.append(torch.mean(g, dim=0))
        return average_gradients

    def SGD_update(
        self,
        partial_gradient_data,
        partial_gradient_model,
    ):
        # Get the correct gradients
        dEdW_data = partial_gradient_data[0]
        dEdW_model = partial_gradient_model[0]
        dEdv_bias_data = partial_gradient_data[1]
        dEdv_bias_model = partial_gradient_model[1]
        dEdh_bias_data = partial_gradient_data[2]
        dEdh_bias_model = partial_gradient_model[2]
        # Gradients
        dW = -dEdW_data + dEdW_model
        dv = -dEdv_bias_data + dEdv_bias_model
        dh = -dEdh_bias_data + dEdh_bias_model
        if self.centering:
            dv = dv - torch.matmul(self.oh, dW)
            dh = dh - torch.matmul(self.ov, dW.t())
        # Add regularization term
        if self.regularization == 'l2':
            dW += self.l2 * 2 * self.W
            dv += self.l2 * 2 * self.v_bias
            dh += self.l2 * 2 * self.h_bias
        elif self.regularization == 'l1':
            dW += self.l1 * torch.sign(self.W)
            dv += self.l1 * torch.sign(self.v_bias)
            dh += self.l1 * torch.sign(self.h_bias)
        # Update parameters in-place
        # and clip
        gnorm = torch.norm(dW) + torch.norm(dv) + torch.norm(dh)
        myclip = (self.lr * 10.0) / gnorm if gnorm > 10 else self.lr
        self.W.add_(myclip * dW)
        self.v_bias.add_(myclip * dv)
        self.h_bias.add_(myclip * dh)

    def Adam_update(
        self,
        t,
        partial_gradient_data,
        partial_gradient_model,
    ):
        # Get the correct gradients
        dEdW_data = partial_gradient_data[0]
        dEdW_model = partial_gradient_model[0]
        dEdv_bias_data = partial_gradient_data[1]
        dEdv_bias_model = partial_gradient_model[1]
        dEdh_bias_data = partial_gradient_data[2]
        dEdh_bias_model = partial_gradient_model[2]
        # Gradients
        dW = -dEdW_data + dEdW_model
        dv = -dEdv_bias_data + dEdv_bias_model
        dh = -dEdh_bias_data + dEdh_bias_model
        if self.centering:
            dv = dv - torch.matmul(self.oh, dW)
            dh = dh - torch.matmul(self.ov, dW.t())
        # Add regularization term
        if self.regularization == 'l2':
            dW += self.l2 * 2 * self.W
            dv += self.l2 * 2 * self.v_bias
            dh += self.l2 * 2 * self.h_bias
        elif self.regularization == 'l1':
            dW += self.l1 * torch.sign(self.W)
            dv += self.l1 * torch.sign(self.v_bias)
            dh += self.l1 * torch.sign(self.h_bias)
        # momentum beta1
        self.m_dW = self.beta1 * self.m_dW + (1 - self.beta1) * dW
        self.m_dv = self.beta1 * self.m_dv + (1 - self.beta1) * dv
        self.m_dh = self.beta1 * self.m_dh + (1 - self.beta1) * dh
        # momentum beta2
        self.v_dW = self.beta2 * self.v_dW + (1 - self.beta2) * (dW**2)
        self.v_dv = self.beta2 * self.v_dv + (1 - self.beta2) * (dv**2)
        self.v_dh = self.beta2 * self.v_dh + (1 - self.beta2) * (dh**2)
        # bias correction
        m_dW_corr = self.m_dW / (1 - self.beta1**t)
        m_dv_corr = self.m_dv / (1 - self.beta1**t)
        m_dh_corr = self.m_dh / (1 - self.beta1**t)
        v_dW_corr = self.v_dW / (1 - self.beta2**t)
        v_dv_corr = self.v_dv / (1 - self.beta2**t)
        v_dh_corr = self.v_dh / (1 - self.beta2**t)
        # Update
        self.W = self.W + self.lr * (m_dW_corr /
                                     (torch.sqrt(v_dW_corr) + self.epsilon))
        self.v_bias = self.v_bias + self.lr * (
            m_dv_corr / (torch.sqrt(v_dv_corr) + self.epsilon))
        self.h_bias = self.h_bias + self.lr * (
            m_dh_corr / (torch.sqrt(v_dh_corr) + self.epsilon))

    def reconstruct(self, data, k, beta=None):
        if beta is None:
            beta = self.model_beta
        self.tstep_keepup()
        data = torch.Tensor(data).to(self.device).to(self.mytype)
        v_model = self.forward(data, k)
        return data.detach().cpu().numpy(), v_model.detach().cpu().numpy()

    def generate(self,
                 n_samples,
                 k,
                 h_binarized=True,
                 from_visible=True,
                 beta=None):
        if beta is None:
            beta = self.model_beta
        self.tstep_keepup()
        if from_visible:
            v = torch.randint(
                high=2,
                size=(n_samples, self.n_visible),
                device=self.device,
                dtype=self.mytype,
            )
        else:
            if h_binarized:
                h = torch.randint(
                    high=2,
                    size=(n_samples, self.n_hidden),
                    device=self.device,
                    dtype=self.mytype,
                )
            else:
                h = torch.rand(n_samples,
                               self.n_hidden,
                               device=self.device,
                               dtype=self.mytype)
            _, v = self.h_to_v(h)
        v_model = self.forward(v, k, beta)
        return v_model.detach().cpu().numpy()

    def clip_weights(self):
        self.W = torch.clip(self.W, self.min_W, self.max_W)
        self.W_t = self.W.t()

    def clip_bias(self):
        self.v_bias = torch.clip(self.v_bias, self.min_W, self.max_W)
        self.h_bias = torch.clip(self.h_bias, self.min_W, self.max_W)

    def _prob_h_given_v(self, v, beta=None):
        if beta is None:
            beta = self.model_beta
        if self.energy_type == 'linear_circuit_5':
            # Circuit minimization
            sol = self._circuit_minimization_visible_clamped(v)
            # h = sol[:,self.h_indices_plus]
            h = (sol[:, self.h_indices_plus] + 1 -
                 sol[:, self.h_indices_minus]) / 2
            # The probability is the outcome of the circuit
            # convert to pytorch tensor
            h = torch.Tensor(h).to(self.device).to(self.mytype)
            return torch.sigmoid(beta * h)
        else:
            return torch.sigmoid(beta * self.delta_eh(v))

    def _prob_v_given_h(self, h, beta=None):
        if beta is None:
            beta = self.model_beta
        if self.energy_type == 'linear_circuit_5':
            # Circuit minimization
            sol = self._circuit_minimization_hidden_clamped(h)
            # v = sol[:,self.v_indices_plus]
            v = (sol[:, self.v_indices_plus] + 1 -
                 sol[:, self.v_indices_minus]) / 2
            # The probability is the outcome of the circuit
            # convert to pytorch tensor
            v = torch.Tensor(v).to(self.device).to(self.mytype)
            return torch.sigmoid(beta * v)
        return torch.sigmoid(beta * self.delta_ev(h))

    def delta_eh(self, v):
        if self.energy_type == 'hopfield':
            return self._delta_eh_hopfield(v)
        elif self.energy_type == 'linear_circuit_4':
            return self._delta_eh_linear_circuit_4(v)

    def delta_ev(self, h):
        if self.energy_type == 'hopfield':
            return self._delta_ev_hopfield(h)
        elif self.energy_type == 'linear_circuit_4':
            return self._delta_ev_linear_circuit_4(h)

    # **** Hopfield transfer functions
    def _delta_eh_hopfield(self, v):
        return torch.mm(v, self.W_t) + self.h_bias

    def _delta_ev_hopfield(self, h):
        return torch.mm(h, self.W) + self.v_bias

    # **** linear_circuit_4 transfer functions
    def _delta_ev_linear_circuit_4(self, h):
        return torch.mm(torch.mul(h, 2) - 1, self._k_diff) + self._vb_posneg

    def _delta_eh_linear_circuit_4(self, v):
        return torch.mm(torch.mul(v, 2) - 1, self._k_diff_t) + self._hb_posneg

    def _update_circuit_variables(self):
        self._k_pos = self.k_pos()
        self._k_neg = self.k_neg()
        self._v_bias_pos = self.v_bias_pos()
        self._v_bias_neg = self.v_bias_neg()
        self._h_bias_pos = self.h_bias_pos()
        self._h_bias_neg = self.h_bias_neg()
        self._k_diff = self._k_pos - self._k_neg
        self._k_diff_t = self._k_diff.t()
        self._v_bias_pos_diff = self._v_bias_pos * (0.5 - self.ground_v)
        self._v_bias_neg_diff = self._v_bias_neg * (0.5 - self.ground_v)
        self._h_bias_pos_diff = self._h_bias_pos * (0.5 - self.ground_h)
        self._h_bias_neg_diff = self._h_bias_neg * (0.5 - self.ground_h)
        self._vb_posneg = -self._v_bias_pos_diff + self._v_bias_neg_diff
        self._hb_posneg = -self._h_bias_pos_diff + self._h_bias_neg_diff

    def derivatives(self, v, h):
        if self.energy_type == 'hopfield':
            return self.derivatives_hopfield(v, h)
        elif (self.energy_type == 'linear_circuit_4'
              or self.energy_type == 'linear_circuit_5'):
            return self.derivatives_linear_circuit_4(v, h)

    # CHANGE THIS FUNCTION TO USE THE CORRECT FREE ENERGY
    def free_energy(self, v, beta=None):
        if beta is None:
            beta = self.model_beta
        if self.energy_type == 'hopfield':
            return self._free_energy_hopfield(v, beta)
        elif self.energy_type == 'linear_circuit_4':
            return self._free_energy_hopfield(v)

    def energy(self, v, h):
        if self.energy_type == 'hopfield':
            return self._energy_hopfield(v, h)
        elif self.energy_type == 'linear_circuit_4':
            return self._energy_linear_circuit_4(v, h)

    def derivatives_hopfield(self, v, h):
        # h has shape (N, n_h) and v has shape (N, n_v), we want result to have shape (N, n_h, n_v)
        if self.centering:
            dEdW = -torch.einsum('ij,ik->ijk', h - self.oh, v - self.ov)
        else:
            dEdW = -torch.einsum('ij,ik->ijk', h, v)
        dEdv_bias = -v
        dEdh_bias = -h
        return (dEdW, dEdv_bias, dEdh_bias)

    def derivatives_linear_circuit_4(self, v, h):
        v_square = torch.square(v).unsqueeze(1)
        vneg_square = torch.square(1 - v).unsqueeze(1)
        h_square = torch.square(h).unsqueeze(2)
        hneg_square = torch.square(1 - h).unsqueeze(2)
        dEdk_pos = (-torch.einsum('ij,ik->ijk', h, v) + 0.5 * v_square + 0.5 *
                    h_square) + (-torch.einsum('ij,ik->ijk', 1 - h, 1 - v) +
                                 0.5 * vneg_square + 0.5 * hneg_square)
        dEdk_neg = (-torch.einsum('ij,ik->ijk', h, 1 - v) + 0.5 * vneg_square +
                    0.5 * h_square) + (-torch.einsum('ij,ik->ijk', 1 - h, v) +
                                       0.5 * v_square + 0.5 * hneg_square)
        dkdw_pos = self.dk_pos().unsqueeze(0)
        dkdw_neg = self.dk_neg().unsqueeze(0)
        dEdW = dEdk_pos * dkdw_pos + dEdk_neg * dkdw_neg
        dEdv_bias_pos = -self.ground_v * v + 0.5 * torch.square(v)
        dEdv_bias_neg = -self.ground_v * (1 - v) + 0.5 * torch.square(1 - v)
        dv_biasdv_pos = self.dv_bias_pos().unsqueeze(0)
        dv_biasdv_neg = self.dv_bias_neg().unsqueeze(0)
        dEdv_bias = dEdv_bias_pos * dv_biasdv_pos + dEdv_bias_neg * dv_biasdv_neg
        dEdh_bias_pos = -self.ground_h * h + 0.5 * torch.square(h)
        dEdh_bias_neg = -self.ground_h * (1 - h) + 0.5 * torch.square(1 - h)
        dh_biasdh_pos = self.dh_bias_pos().unsqueeze(0)
        dh_biasdh_neg = self.dh_bias_neg().unsqueeze(0)
        dEdh_bias = dEdh_bias_pos * dh_biasdh_pos + dEdh_bias_neg * dh_biasdh_neg
        return (dEdW, dEdv_bias, dEdh_bias)

    def _circuit_minimization_visible_clamped(self, v):
        # # convert weights to conductances
        # self._weights_to_conductances()
        # # set the conductances in the same order as they are internally stored
        # conductances = [self.circuit.graph.edges[e]['conductance'] for e in self.circuit.graph.edges]
        # self.circuit.setConductances(conductances)
        f = np.concatenate(
            (v, 1 - v, np.tile([self.ground_v, self.ground_h],
                               (v.shape[0], 1))),
            axis=1)
        return np.array(
            [self.circuit.solve(self.Q_visible_clamped, fi) for fi in f])

    def _circuit_minimization_hidden_clamped(self, h):
        # # convert weights to conductances
        # self._weights_to_conductances()
        # # set the conductances in the same order as they are internally stored
        # conductances = [self.circuit.graph.edges[e]['conductance'] for e in self.circuit.graph.edges]
        # self.circuit.setConductances(conductances)
        f = np.concatenate(
            (h, 1 - h, np.tile([self.ground_v, self.ground_h],
                               (h.shape[0], 1))),
            axis=1)
        return np.array(
            [self.circuit.solve(self.Q_hidden_clamped, fi) for fi in f])

    def _weights_to_conductances(self):
        """Function to transform weights to conductances. Used for power
        minimization in the real circuits.

        Returns:
            Conductances of the circuit.
        """
        k_plus = self.k_pos().ravel()
        k_minus = self.k_neg().ravel()
        q_plus = self.v_bias_pos()
        q_minus = self.v_bias_neg()
        c_plus = self.h_bias_pos()
        c_minus = self.h_bias_neg()
        # update the conductances
        # # USE UPDATE FUNCTION INSTEAD OF ADD_EDGE
        # for i in range(self.n_visible):
        #     for j in range(self.n_hidden):
        #         # interactions between visible and hidden units
        #         # self.circuit.add_edge(f'$v_{i}^+$', f'$h_{j}^+$', conductance=k_plus[j][i])
        #         self.circuit.graph.edges[f'$v_{i}^+$',f'$h_{j}^+$']['conductance'] = k_plus[j][i]
        #         self.circuit.graph.edges[f'$v_{i}^+$',f'$h_{j}^-$']['conductance'] = k_minus[j][i]
        #         self.circuit.graph.edges[f'$v_{i}^-$',f'$h_{j}^+$']['conductance'] = k_minus[j][i]
        #         self.circuit.graph.edges[f'$v_{i}^-$',f'$h_{j}^-$']['conductance'] = k_plus[j][i]

        #         # self.circuit.add_edge(f'$v_{i}^+$', f'$h_{j}^-$', conductance=k_minus[j][i])
        #         # self.circuit.add_edge(f'$v_{i}^-$', f'$h_{j}^+$', conductance=k_minus[j][i])
        #         # self.circuit.add_edge(f'$v_{i}^-$', f'$h_{j}^-$', conductance=k_plus[j][i])
        #     # interactions between visible and ground_visible
        #     # self.circuit.add_edge(f'$v_{i}^+$', r'$g_v$', conductance=q_plus[i])
        #     # self.circuit.add_edge(f'$v_{i}^-$', r'$g_v$', conductance=q_minus[i])
        #     self.circuit.graph.edges[f'$v_{i}^+$',r'$g_v$']['conductance'] = q_plus[i]
        #     self.circuit.graph.edges[f'$v_{i}^-$',r'$g_v$']['conductance'] = q_minus[i]

        # for j in range(self.n_hidden):
        #     # interactions between hidden and ground_hidden
        #     # self.circuit.add_edge(r'$g_h$', f'$h_{j}^+$', conductance=c_plus[j])
        #     # self.circuit.add_edge(r'$g_h$', f'$h_{j}^-$', conductance=c_minus[j])
        #     self.circuit.graph.edges[r'$g_h$',f'$h_{j}^+$']['conductance'] = c_plus[j]
        #     self.circuit.graph.edges[r'$g_h$',f'$h_{j}^-$']['conductance'] = c_minus[j]

        # Concatenate all the conductances according to the edge_indexmap
        conductances = np.concatenate(
            (k_plus, k_plus, k_minus, k_minus, q_plus, q_minus, c_plus,
             c_minus))[self.edge_indexmap]
        # if conductances are too small, set them to a minimum value
        conductances[conductances <
                     self.min_conductance] = self.min_conductance
        return conductances

    # ***** Get high-level circuit weights
    def k_pos(self):
        return torch.relu(self.W)

    def k_neg(self):
        return torch.relu(-self.W)

    def dk_pos(self):
        return (self.W > 0).to(self.W.dtype)

    def dk_neg(self):
        return -(self.W < 0).to(self.W.dtype)

    def v_bias_pos(self):
        return torch.relu(self.v_bias)

    def v_bias_neg(self):
        return torch.relu(-self.v_bias)

    def dv_bias_pos(self):
        return (self.v_bias > 0).to(self.v_bias.dtype)

    def dv_bias_neg(self):
        return -(self.v_bias < 0).to(self.v_bias.dtype)

    # ***** Get high-level circuit bias
    def h_bias_pos(self):
        return torch.relu(self.h_bias)

    def h_bias_neg(self):
        return torch.relu(-self.h_bias)

    def dh_bias_pos(self):
        return (self.h_bias > 0).to(self.h_bias.dtype)

    def dh_bias_neg(self):
        return -(self.h_bias < 0).to(self.h_bias.dtype)

    # **** PLOTTING
    def plot_weights(self, t):
        Ndata = self.W.shape[0]
        # Reshape the matrix into a 3D array
        data_3d = self.W.detach().cpu().numpy().reshape(Ndata, 28, 28)
        # Determine the number of rows and columns for the subplot grid
        num_rows = int(np.ceil(np.sqrt(Ndata)))
        num_cols = int(np.ceil(Ndata / num_rows))
        # Create a figure and axis for the plot
        fig, ax = plt.subplots(nrows=num_rows,
                               ncols=num_cols,
                               figsize=(10, 10))
        # Iterate over the submatrices and plot them
        for i in range(Ndata):
            row = i // num_cols
            col = i % num_cols
            ax[row, col].imshow(data_3d[i], cmap='magma')
            ax[row, col].axis('off')
        # Remove empty subplots if the number of submatrices doesn't fill the entire grid
        if num_rows * num_cols > Ndata:
            for i in range(Ndata, num_rows * num_cols):
                row = i // num_cols
                col = i % num_cols
                fig.delaxes(ax[row, col])
        # Adjust the spacing between subplots
        plt.suptitle('Weights epoch {}'.format(t))
        plt.subplots_adjust(wspace=0.05, hspace=0.05, top=0.9)
        # Get the minimum and maximum values from the data
        vmin = np.min(self.W.detach().cpu().numpy())
        vmax = np.max(self.W.detach().cpu().numpy())
        # Create a dummy image for the colorbar
        dummy_img = np.zeros((1, 1))  # Dummy image with all zeros
        # Add a colorbar using the dummy image as the mappable
        cax = fig.add_axes([0.93, 0.15, 0.02, 0.7])  # Position of the colorbar
        plt.colorbar(plt.imshow(dummy_img, cmap='magma', vmin=vmin, vmax=vmax),
                     cax=cax)
        # Adjust the height of the colorbar axes to match the height of the figure
        cax.set_aspect('auto')

    # ** Plotting bias
    def plot_bias(self, t):
        h_bias = self.h_bias.detach().cpu().numpy()
        v_bias = self.v_bias.detach().cpu().numpy()
        # Set up the figure with two subplots
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 6))
        # Plot histogram for hidden biases
        ax1.hist(h_bias, bins=20, color='blue', edgecolor='black')
        ax1.set_xlabel('Values')
        ax1.set_ylabel('Frequency')
        ax1.set_title('Hidden Biases epoch {}'.format(t))
        # Plot histogram for visible biases
        ax2.hist(v_bias, bins=20, color='red', edgecolor='black')
        ax2.set_xlabel('Values')
        ax2.set_ylabel('Frequency')
        ax2.set_title('Visible Biases epoch {}'.format(t))
        # Adjust layout for better readability
        plt.tight_layout()

    def _center(self, g_v=0.5, g_h=0.5):
        W_centered = self.W_t
        v_bias_centered = (self.v_bias + 0.5 * W_centered.sum(dim=1)) / g_v
        h_bias_centered = (self.h_bias + 0.5 * W_centered.sum(dim=0)) / g_h

        return W_centered, v_bias_centered, h_bias_centered

    def _RKM_v_to_h(self,
                    v_centered,
                    W_centered,
                    v_bias_centered,
                    h_bias_centered,
                    g_v=0.5,
                    g_h=0.5):
        # Compute the equilibrated hidden units
        h_eq = (torch.mm(v_centered, W_centered) + g_h * h_bias_centered) / (
            (torch.abs(W_centered).sum(dim=0) + torch.abs(h_bias_centered)))
        return h_eq

    def _RKM_h_to_v(self,
                    h_centered,
                    W_centered,
                    v_bias_centered,
                    h_bias_centered,
                    g_v=0.5,
                    g_h=0.5):
        # Compute the equilibrated visible units
        v_eq = (torch.mm(h_centered, W_centered.T) + g_v * v_bias_centered) / (
            torch.abs(W_centered).sum(dim=1) + torch.abs(v_bias_centered))
        return v_eq

    def power_forward(self, v, g_v=0.5, g_h=0.5):
        """Computes the power dissipated by the RKM in the forward pass.

        Args:
            v: visible units, shape (N, n_v)
        Returns:
            Power dissipated by the RKM, shape (N,)
        """
        v_centered = v - 0.5
        W_centered, v_bias_centered, h_bias_centered = self._center(g_v, g_h)
        h_eq = self._RKM_v_to_h(v_centered, W_centered, v_bias_centered,
                                h_bias_centered, g_v, g_h)

        # Compute the power dissipated by the RKM
        power = (torch.matmul(v_centered**2,
                              torch.abs(W_centered / 2).sum(dim=1)) +
                 torch.matmul(h_eq**2,
                              torch.abs(W_centered / 2).sum(dim=0)) -
                 torch.einsum('ij,ji->i', v_centered,
                              torch.matmul(W_centered, h_eq.T)) +
                 torch.matmul((h_eq**2 + g_h**2), torch.abs(h_bias_centered)) -
                 torch.matmul(h_eq, h_bias_centered) * g_h)

        return power

    def power_backward(self, h, g_v=0.5, g_h=0.5):
        """Computes the power dissipated by the RKM in the backward pass.

        Args:
            h: hidden units, shape (N, n_h)
        Returns:
            Power dissipated by the RKM, shape (N,)
        """
        h_centered = h - 0.5
        W_centered, v_bias_centered, h_bias_centered = self._center(g_v, g_h)
        v_eq = self._RKM_h_to_v(h_centered, W_centered, v_bias_centered,
                                h_bias_centered, g_v, g_h)

        # Compute the power dissipated by the RKM
        power = (torch.matmul(h_centered**2,
                              torch.abs(W_centered / 2).sum(dim=0)) +
                 torch.matmul(v_eq**2,
                              torch.abs(W_centered / 2).sum(dim=1)) -
                 torch.einsum('ij,ji->i', h_centered,
                              torch.matmul(W_centered.T, v_eq.T)) +
                 torch.matmul((v_eq**2 + g_v**2), torch.abs(v_bias_centered)) -
                 torch.matmul(v_eq, v_bias_centered) * g_v)

        return power

    def av_power_forward(self, v, g_v=0.5, g_h=0.5):
        """Computes the average power dissipated by the RKM in the forward
        pass.

        Args:
            v: visible units, shape (N, n_v)
        Returns:
            Average power dissipated by the RKM
        """
        return self.power_forward(v, g_v, g_h).mean()

    def av_power_backward(self, h, g_v=0.5, g_h=0.5):
        """Computes the average power dissipated by the RKM in the backward
        pass.

        Args:
            h: hidden units, shape (N, n_h)
        Returns:
            Average power dissipated by the RKM
        """
        return self.power_backward(h, g_v, g_h).mean()

    def relaxation_times(self):
        """Computes the relaxation times of the RKM in the forward pass.

        Args:
            None
        Returns:
            t_forward: relaxation times of the RKM in the forward pass, shape (n_v,)
            t_backward: relaxation times of the RKM in the backward pass, shape (n_h,)
        """
        W_centered, v_bias_centered, h_bias_centered = self._center()
        t_forward = 1 / (torch.abs(W_centered / 2).sum(dim=0) +
                         torch.abs(h_bias_centered))
        t_backward = 1 / (torch.abs(W_centered / 2).sum(dim=1) +
                          torch.abs(v_bias_centered))

        return t_forward, t_backward



import unittest
import torch
from pyrkm import RBM


class TestRBM(unittest.TestCase):

    def setUp(self):
        self.rbm = RBM('test_model', n_vis=10, n_hin=5)

    def test_init(self):
        self.assertEqual(self.rbm.name, 'test_model')
        self.assertEqual(self.rbm.n_visible, 10)
        self.assertEqual(self.rbm.n_hidden, 5)
        self.assertIsInstance(self.rbm.W, torch.Tensor)
        self.assertIsInstance(self.rbm.v_bias, torch.Tensor)
        self.assertIsInstance(self.rbm.h_bias, torch.Tensor)

    def test_weights(self):
        self.assertEqual(self.rbm.W.shape, (5, 10))
        self.assertEqual(self.rbm.v_bias.shape, (10, ))
        self.assertEqual(self.rbm.h_bias.shape, (5, ))

    def test_device(self):
        self.assertEqual(
            self.rbm.device,
            torch.device('cuda' if torch.cuda.is_available() else 'cpu'),
        )

    def test_model_beta(self):
        self.assertEqual(self.rbm.model_beta, 1)

    def test_regularization(self):
        self.assertFalse(self.rbm.regularization)
        self.assertFalse(hasattr(self.rbm, 'l1'))
        self.assertFalse(hasattr(self.rbm, 'l2'))

    def test_optimizer(self):
        self.assertEqual(self.rbm.optimizer, 'Adam')
        self.assertEqual(self.rbm.beta1, 0.9)
        self.assertEqual(self.rbm.beta2, 0.999)
        self.assertEqual(self.rbm.epsilon, 1e-8)

    def test_weights_initialization(self):
        self.assertTrue(
            torch.all(torch.logical_and(self.rbm.W >= -100, self.rbm.W
                                        <= 100)))
        self.assertTrue(
            torch.all(
                torch.logical_and(self.rbm.h_bias >= -100, self.rbm.h_bias
                                  <= 100)))


if __name__ == '__main__':
    unittest.main()



import torch
import torch.nn.init as init

from .rbm_pytorch import RBM


################################################################
# Class for the fitlering RBM model
class filtering_RBM(RBM):

    def __init__(
        self,
        model_name,
        n_vis=784,
        n_hin=50,
        k=1,
        lr=0.01,
        max_epochs=200000,
        minMax_W=(-100, 100),
        energy_type='hopfield',
        optimizer='Adam',
        regularization=False,
        l1_factor=1e-7,
        l2_factor=1e-7,
        ground_v=1,
        ground_h=1,
        batch_size=64,
        train_algo='CD',
        centering=False,
        average_data=None,
        sampling_model_beta=1,
        use_mask=False,
        mytype=torch.double,
        is_conditional=False,
    ):
        # Initialize base class parameters
        super(filtering_RBM, self).__init__(
            model_name,
            n_vis,
            n_vis,
            k,
            lr,
            max_epochs,
            minMax_W,
            energy_type,
            optimizer,
            regularization,
            l1_factor,
            l2_factor,
            ground_v,
            ground_h,
            batch_size,
            train_algo,
            centering,
            average_data,
            sampling_model_beta,
            mytype,
        )
        # Initialize new parameters
        # If true, add conditional layer
        self.is_conditional = is_conditional
        # Initialize weights
        if is_conditional:
            self.D = torch.randn(
                (
                    n_hin,
                    n_vis,
                ),
                dtype=self.mytype,
                device=self.device,
            )
            self.D = self.D.contiguous()
            init.xavier_normal_(self.D)

        # use mask for recommendation systems
        self.use_mask = use_mask

    # Function to average the gradients, excluding the masked elements
    def maskedmean(self, x):
        mask_sum = (x != 0).to(torch.int).sum(dim=0)

        # Check if any element in mask_sum is zero
        if (mask_sum == 0).any():
            # print("Warning: Zero elements in mask_sum. Replacing with 1.")
            # Replace zero elements with 1 in mask_sum
            mask_sum = torch.where(mask_sum == 0, 1, mask_sum)

        return torch.div(x.sum(dim=0), mask_sum)

    def tstep_keepup(self):
        if self.energy_type != 'hopfield':
            self.clip_weights()
            self.clip_bias()
            self._update_circuit_variables(self.is_conditional)
        else:
            self.W_t = self.W.t()
        if self.is_conditional:
            self.D_t = self.D.t()

    def _update_circuit_variables(self):
        self._k_pos = self.k_pos()
        self._k_neg = self.k_neg()
        self._v_bias_pos = self.v_bias_pos()
        self._v_bias_neg = self.v_bias_neg()
        self._h_bias_pos = self.h_bias_pos()
        self._h_bias_neg = self.h_bias_neg()
        self._k_diff = self._k_pos - self._k_neg
        self._k_diff_t = self._k_diff.t()
        self._v_bias_pos_diff = self._v_bias_pos * (0.5 - self.ground_v)
        self._v_bias_neg_diff = self._v_bias_neg * (0.5 - self.ground_v)
        self._h_bias_pos_diff = self._h_bias_pos * (0.5 - self.ground_h)
        self._h_bias_neg_diff = self._h_bias_neg * (0.5 - self.ground_h)
        self._vb_posneg = -self._v_bias_pos_diff + self._v_bias_neg_diff
        self._hb_posneg = -self._h_bias_pos_diff + self._h_bias_neg_diff
        if self.is_conditional:
            self._D_pos = self.D_pos()
            self._D_neg = self.D_neg()
            self._D_diff = self._D_pos - self._D_neg
            self._D_diff_t = self._D_diff.t()

    def SGD_update(
        self,
        partial_gradient_data,
        partial_gradient_model,
    ):
        # Get the correct gradients
        dEdW_data = partial_gradient_data[0]
        dEdW_model = partial_gradient_model[0]
        dEdv_bias_data = partial_gradient_data[1]
        dEdv_bias_model = partial_gradient_model[1]
        dEdh_bias_data = partial_gradient_data[2]
        dEdh_bias_model = partial_gradient_model[2]
        # Gradients
        dW = -dEdW_data + dEdW_model
        dv = -dEdv_bias_data + dEdv_bias_model
        dh = -dEdh_bias_data + dEdh_bias_model
        if self.is_conditional:
            dEdD_data = partial_gradient_data[3]
            dEdD_model = partial_gradient_model[3]
            dD = -dEdD_data + dEdD_model
        if self.centering:
            dv = dv - torch.matmul(self.oh, dW)
            dh = dh - torch.matmul(self.ov, dW.t())
        # Add regularization term
        if self.regularization == 'l2':
            dW += self.l2 * 2 * self.W
            dv += self.l2 * 2 * self.v_bias
            dh += self.l2 * 2 * self.h_bias
        elif self.regularization == 'l1':
            dW += self.l1 * torch.sign(self.W)
            dv += self.l1 * torch.sign(self.v_bias)
            dh += self.l1 * torch.sign(self.h_bias)
        # Update parameters in-place
        # and clip
        if self.is_conditional:
            gnorm = torch.norm(dW) + torch.norm(dv) + torch.norm(
                dh) + torch.norm(dD)
        else:
            gnorm = torch.norm(dW) + torch.norm(dv) + torch.norm(dh)
        myclip = (self.lr * 10.0) / gnorm if gnorm > 10 else self.lr
        self.W.add_(myclip * dW)
        self.v_bias.add_(myclip * dv)
        self.h_bias.add_(myclip * dh)
        if self.is_conditional:
            self.D.add_(self.lr * dD * myclip)

    def delta_eh(self, v):
        if self.energy_type == 'hopfield':
            if self.use_mask:
                return self._delta_eh_hopfield_masked(v)
            else:
                return self._delta_eh_hopfield(v)
        elif self.energy_type == 'linear_circuit_4':
            if self.use_mask:
                return self._delta_eh_linear_circuit_4_masked(v)
            else:
                return self._delta_eh_linear_circuit_4(v)

    def delta_ev(self, h):
        if self.energy_type == 'hopfield':
            if self.use_mask:
                return self._delta_ev_hopfield_masked(h)
            else:
                return self._delta_ev_hopfield(h)
        elif self.energy_type == 'linear_circuit_4':
            if self.use_mask:
                return self._delta_ev_linear_circuit_4_masked(h)
            else:
                return self._delta_ev_linear_circuit_4(h)

    # **** Hopfield transfer functions
    def _delta_eh_hopfield(self, v):
        return torch.mm(v, self.W_t) + self.h_bias

    def _delta_ev_hopfield(self, h):
        return torch.mm(h, self.W) + self.v_bias

    # **** (masked) Hopfield transfer functions
    def _delta_eh_hopfield_masked(self, v):
        #        Wtm = self.W_t *self.mask.unsqueeze(-1)
        #        return torch.bmm(v.unsqueeze(1), Wtm).squeeze(1) + self.h_bias + torch.mm(self.mask , self.D_t)
        if self.is_conditional:
            return (torch.mm(v * self.mask, self.W_t) + self.h_bias +
                    torch.mm(self.mask, self.D_t))
        else:
            return torch.mm(v * self.mask, self.W_t) + self.h_bias

    def _delta_ev_hopfield_masked(self, h):
        return torch.mm(h, self.W) + self.v_bias

    # **** linear_circuit_4 transfer functions
    def _delta_ev_linear_circuit_4(self, h):
        return torch.mm(torch.mul(h, 2) - 1, self._k_diff) + self._vb_posneg

    def _delta_eh_linear_circuit_4(self, v):
        return torch.mm(torch.mul(v, 2) - 1, self._k_diff_t) + self._hb_posneg

    # **** (masked) linear_circuit_4 transfer functions
    def _delta_ev_linear_circuit_4_masked(self, h):
        return torch.mm(torch.mul(h, 2) - 1, self._k_diff) + self._vb_posneg

    def _delta_eh_linear_circuit_4_masked(self, v):
        if self.is_conditional:
            return (torch.mm(torch.mul(v * self.mask, 2) - 1, self._k_diff_t) +
                    self._hb_posneg +
                    torch.mm(torch.mul(self.mask, 2) - 1, self._D_diff_t))
        else:
            return (torch.mm(torch.mul(v * self.mask, 2) - 1, self._k_diff_t) +
                    self._hb_posneg)

    def derivatives_hopfield(self, v, h):
        # h has shape (N, n_h) and v has shape (N, n_v), we want result to have shape (N, n_h, n_v)
        if self.centering:
            dEdW = -torch.einsum('ij,ik->ijk', h - self.oh, v - self.ov)
        else:
            dEdW = -torch.einsum('ij,ik->ijk', h, v * self.mask)
        dEdv_bias = -v * self.mask
        dEdh_bias = -h
        if self.is_conditional:
            dEdD = -torch.einsum('ij,ik->ijk', h, self.mask)
        # They have to be renormalized according to the size of the mask
        # masksizefactor = self.n_visible/self.mask.sum(-1)
        if self.is_conditional:
            return (dEdW, dEdv_bias, dEdh_bias, dEdD)
            # return dEdW*masksizefactor.unsqueeze(1).unsqueeze(1), dEdv_bias*masksizefactor.unsqueeze(1), dEdh_bias, dEdD#*masksizefactor.unsqueeze(1).unsqueeze(1)
        else:
            return (dEdW, dEdv_bias, dEdh_bias, None)
            # return dEdW*masksizefactor.unsqueeze(1).unsqueeze(1), dEdv_bias*masksizefactor.unsqueeze(1), dEdh_bias, None

    def derivatives_linear_circuit_4(self, v, h):
        torch.square(v).unsqueeze(1)
        torch.square(1 - v).unsqueeze(1)
        negh = 1 - h
        h_square = torch.square(h).unsqueeze(2)
        hneg_square = torch.square(negh).unsqueeze(2)
        v_masked = v * self.mask
        negv_masked = 1 - v_masked
        v_square_masked = torch.square(v_masked).unsqueeze(1)
        vneg_square_masked = torch.square(1 - v_masked).unsqueeze(1)

        ein_h_vm = torch.einsum('ij,ik->ijk', h, v_masked)
        ein_negh_vm = torch.einsum('ij,ik->ijk', negh, v_masked)
        ein_h_negvm = torch.einsum('ij,ik->ijk', h, negv_masked)
        ein_negh_negvm = torch.einsum('ij,ik->ijk', negh, negv_masked)

        dEdk_pos = 0.5 * (
            (-2 * ein_h_vm + v_square_masked + h_square) +
            (-2 * ein_negh_negvm + vneg_square_masked + hneg_square))
        dEdk_neg = 0.5 * ((-2 * ein_h_negvm + vneg_square_masked + h_square) +
                          (-2 * ein_negh_vm + v_square_masked + hneg_square))
        dkdw_pos = self.dk_pos().unsqueeze(0)
        dkdw_neg = self.dk_neg().unsqueeze(0)
        dEdW = dEdk_pos * dkdw_pos + dEdk_neg * dkdw_neg

        dEdv_bias_pos = -self.ground_v * v_masked + 0.5 * torch.square(
            v_masked)
        dEdv_bias_neg = -self.ground_v * negv_masked + 0.5 * torch.square(
            negv_masked)
        dv_biasdv_pos = self.dv_bias_pos().unsqueeze(0)
        dv_biasdv_neg = self.dv_bias_neg().unsqueeze(0)
        dEdv_bias = dEdv_bias_pos * dv_biasdv_pos + dEdv_bias_neg * dv_biasdv_neg

        dEdh_bias_pos = -self.ground_h * h + 0.5 * torch.square(h)
        dEdh_bias_neg = -self.ground_h * (negh) + 0.5 * torch.square(negh)
        dh_biasdh_pos = self.dh_bias_pos().unsqueeze(0)
        dh_biasdh_neg = self.dh_bias_neg().unsqueeze(0)
        dEdh_bias = dEdh_bias_pos * dh_biasdh_pos + dEdh_bias_neg * dh_biasdh_neg

        if self.is_conditional:
            ## No this gradient is wrong! (this is same shape as W, but it is not like this in C-RBM)
            negm = 1 - self.mask
            m_square = torch.square(self.mask).unsqueeze(1)
            mneg_square = torch.square(negm).unsqueeze(1)

            ein_h_m = torch.einsum('ij,ik->ijk', h, self.mask)
            ein_h_negm = torch.einsum('ij,ik->ijk', h, negm)
            ein_negh_m = torch.einsum('ij,ik->ijk', negh, self.mask)
            ein_negh_negm = torch.einsum('ij,ik->ijk', negh, negm)

            dEdD_pos = 0.5 * ((-2 * ein_h_m + m_square + h_square) +
                              (-2 * ein_negh_negm + mneg_square + hneg_square))
            dEdD_neg = 0.5 * ((-2 * ein_h_negm + mneg_square + h_square) +
                              (-2 * ein_negh_m + m_square + hneg_square))
            dDdw_pos = self.dD_pos().unsqueeze(0)
            dDdw_neg = self.dD_neg().unsqueeze(0)
            dEdD = dEdD_pos * dDdw_pos + dEdD_neg * dDdw_neg

            # negm = 1-self.mask
            # dEdD = torch.einsum('ij,ik->ijk',dEdh_bias_pos*dh_biasdh_pos, self.mask) + torch.einsum('ij,ik->ijk',dEdh_bias_neg*dh_biasdh_neg, negm)

        # They have to be renormalized according to the size of the mask
        # masksizefactor = self.n_visible/self.mask.sum(-1)
        if self.is_conditional:
            return (dEdW, dEdv_bias, dEdh_bias, dEdD)
            # return dEdW*masksizefactor.unsqueeze(1).unsqueeze(1), dEdv_bias*masksizefactor.unsqueeze(1), dEdh_bias, dEdD#*masksizefactor.unsqueeze(1).unsqueeze(1)
        else:
            return (dEdW, dEdv_bias, dEdh_bias, None)
            # return dEdW*masksizefactor.unsqueeze(1).unsqueeze(1), dEdv_bias*masksizefactor.unsqueeze(1), dEdh_bias, None

    # ***** Get high-level conditional weights
    def D_pos(self):
        return torch.relu(self.D)

    def D_neg(self):
        return torch.relu(-self.D)

    def dD_pos(self):
        return (self.D > 0).to(self.D.dtype)

    def dD_neg(self):
        return -(self.D < 0).to(self.D.dtype)



import numpy as np
import torch

# ********* Functions to compute AIS score
# they do not work as expected (different results from literature)
# Overall, since we can not compute AIS for circuit models or ReLU approach, it is not worth to debug this

# from old implementation:
# *** For ReLU models the probability is not defined, so free energy indicators are ill defined
# ll2 = ComputeFreeEnergyAIS(test_model, v_bias_base.to(test_model.device), nb=1000, nI=1000)
## Calculate logZ using AIS
# logZ = annealed_importance_sampling(test_model,v_bias_base.to(test_model.device),num_chains=1000,betas=1000).detach().cpu()
## Calculate the loglikelihood of this batch of data
# LL = torch.mean(log_likelihood_v(test_model,logZ, train_data.to(test_model.device), v_bias_base.to(test_model.device)))
# ***


def annealed_importance_sampling(
    model,
    v_bias_base,
    num_chains=1000,
    betas=1000,
):
    """Approximates the partition function for the given model using annealed
    importance sampling.

    .. seealso:: Accurate and Conservative Estimates of MRF Log-likelihood using Reverse Annealing \
                 http://arxiv.org/pdf/1412.8566.pdf

    :param model: The model.
    :type model: Valid RBM model.

    :param num_chains: Number of AIS runs.
    :type num_chains: int

    :param k: Number of Gibbs sampling steps.
    :type k: int

    :param betas: Number or a list of inverse temperatures to sample from.
    :type betas: int, numpy array [num_betas]

    :return: | Mean estimated log partition function,
    :rtype: float
    """
    # Setup temerpatures
    betas = torch.linspace(0.0, 1.0, betas)

    ## Sample the first time from the base model (start does not matter since beta=0)
    # _, v = model.h_to_v(torch.zeros((num_chains, model.n_hidden)).to(model.device).to(torch.double), betas[0])
    # _, h = model.v_to_h(v, betas[0])
    v = torch.bernoulli(
        torch.sigmoid(v_bias_base.unsqueeze(0).repeat(num_chains, 1)))
    # _, h = model.v_to_h(v, betas[0])
    # E = model.energy(v,h)
    # Base distro:
    #    P(x)=exp[-beta*(v*v_bias_base)]/[2**nh Zv]
    #    Zv = 1 + exp[beta*v_bias_base]
    # E += torch.mv(v, v_bias_base)

    # Calculate the unnormalized probabilties of v
    lnpvsum = -unnormalized_log_probability_v(model, v, v_bias_base, betas[0],
                                              True)

    for beta in betas[1:betas.shape[0] - 1]:
        # Calculate the unnormalized probabilties of v
        lnpvsum += unnormalized_log_probability_v(model, v, v_bias_base, beta,
                                                  True)

        # Sample k times from the intermidate distribution
        for _ in range(model.k):
            _, h = model.v_to_h(v, beta)
            _, v = model.h_to_v(h, beta)

        # Calculate the unnormalized probabilties of v
        lnpvsum -= unnormalized_log_probability_v(model, v, v_bias_base, beta,
                                                  True)

        # E += model.energy(v,h)

    # Calculate the unnormalized probabilties of v
    lnpvsum += unnormalized_log_probability_v(model, v, v_bias_base,
                                              betas[betas.shape[0] - 1], True)

    # Calculate an estimate of logz .
    logz = log_sum_exp(lnpvsum) - np.log(num_chains)

    # Calculate partition function of base distribution (False=uniform; True=biased on data)
    baselogz = _base_log_partition(model, v_bias_base, True)

    # Add the base partition function
    logz = logz + baselogz

    # db = 1/betas.shape[0]
    # d = (-db*E).to(torch.double)

    ## to avoid overflow introduce d0
    # d0 = d.mean()
    # logz = torch.log( torch.mean( torch.exp(d-d0)))+d0

    return logz


def unnormalized_log_probability_v(model,
                                   v,
                                   v_bias_base,
                                   beta=None,
                                   use_base_model=False):
    """Computes the unnormalized log probabilities of v.

    :param v: Visible states.
    :type v: numpy array [batch size, input dim]
    :param beta: Allows to sample from a given inverse temperature beta,
        or if a vector is given to sample from \ different betas
        simultaneously.None is equivalent to pass the value 1.0.
    :type beta: None, float or numpy array [batch size, 1]
    :param use_base_model: If true uses the base model, i.e. the MLE of
        the bias values.
    :type use_base_model: bool
    :return: Unnormalized log probability of v.
    :rtype: numpy array [batch size, 1]
    """
    temp_v = v  # - model.ov
    activation = torch.mm(temp_v, model.W.t()) + model.h_bias
    bias = torch.mv(temp_v, model.v_bias)
    if beta is not None:
        activation *= beta
        bias *= beta
        if use_base_model is True:
            bias += (1.0 - beta) * torch.mv(temp_v, v_bias_base)
    return bias + torch.sum(torch.log(1 + torch.exp(activation)), axis=1)


def _base_log_partition(model, v_bias_base, use_base_model=False):
    """ Returns the base partition function for a given visible bias. .. Note:: that for AIS we need to be able to \
        calculate the partition function of the base distribution exactly. Furthermore it is beneficial if the \
        base distribution is a good approximation of the target distribution. A good choice is therefore the \
        maximum likelihood estimate of the visible bias, given the data.

    :param use_base_model: If true uses the base model, i.e. the MLE of the bias values.
    :type use_base_model: bool

    :return: Partition function for zero parameters.
    :rtype: float
    """
    if use_base_model is True:
        # return torch.sum(torch.log(torch.exp(-model.ov * v_bias_base) + torch.exp((1.0 - model.ov) * v_bias_base)) ) + model.n_hidden * np.log(2.0)
        # return torch.sum(torch.log(1 + torch.exp(v_bias_base)) ) + model.n_hidden * np.log(2.0)
        return torch.sum(
            torch.log(1 +
                      torch.exp(v_bias_base))) + model.n_hidden * np.log(2.0)
    else:
        return model.n_visible * np.log(2.0) + model.n_hidden * np.log(2.0)


def log_likelihood_v(model, logz, data, v_bias_base, beta=None):
    """Computes the log-likelihood (LL) for a given model and visible data
    given its log partition function.

    :Info: logz needs to be the partition function for the same beta
    (i.e. beta = 1.0)!

    :param logz: The logarithm of the partition function.
    :type logz: float
    :param data: The visible data.
    :type data: 2D array [num samples, num input dim]
    :param beta: Inverse temperature(s) for the models energy.
    :type beta: None, float, numpy array [batchsize,1]
    :return: The log-likelihood for each sample.
    :rtype: numpy array [num samples]
    """
    return log_probability_v(model, logz, data, v_bias_base, beta)


def log_probability_v(model,
                      logz,
                      v,
                      v_bias_base,
                      beta=None,
                      use_base_model=False):
    """Computes the log-probability / LogLikelihood(LL) for the given visible
    units for this model. To estimate \ the LL we need to know the logarithm of
    the partition function Z. For small models it is possible to \ calculate Z,
    however since this involves calculating all possible hidden states, it is
    intractable for \ bigger models. As an estimation method annealed
    importance sampling (AIS) can be used instead.

    :param logz: The logarithm of the partition function.
    :type logz: float
    :param v: Visible states.
    :type v: numpy array [batch size, input dim]
    :param beta: Allows to sample from a given inverse temperature beta,
        or if a vector is given to sample from \ different betas
        simultaneously.None is equivalent to pass the value 1.0.
    :type beta: None, float or numpy array [batch size, 1]
    :param use_base_model: If true uses the base model, i.e. the MLE of
        the bias values.
    :type use_base_model: bool
    :return: Log probability for visible_states.
    :rtype: numpy array [batch size, 1]
    """
    return (unnormalized_log_probability_v(model, v, v_bias_base, beta,
                                           use_base_model) - logz)


def log_sum_exp(x, axis=0):
    alpha = x.max(axis).values - np.log(np.finfo(np.float64).max) / 2.0
    if axis == 1:
        return torch.squeeze(
            alpha + torch.log(torch.sum(torch.exp(x.T - alpha), axis=0)))
    else:
        return torch.squeeze(alpha +
                             torch.log(torch.sum(torch.exp(x -
                                                           alpha), axis=0)))


def log_diff_exp(x, axis=0):
    alpha = x.max(axis).values - np.log(np.finfo(np.float64).max) / 2.0
    if axis == 1:
        return torch.squeeze(
            alpha + torch.log(torch.diff(torch.exp(x.T - alpha), n=1, axis=0)))
    else:
        return torch.squeeze(
            alpha + torch.log(torch.diff(torch.exp(x - alpha), n=1, axis=0)))


# Uniform prior
# p0(x,h)= 1/Z0
# Z0 = 2**(nvis+nhid)
def ComputeFreeEnergyAIS(model, v_bias_base, nb, nI):
    blist = torch.arange(0, 1.000001, 1.0 / nb)
    torch.zeros(model.n_visible + model.n_hidden, nI, device=model.device)
    torch.zeros(model.n_hidden, nI, device=model.device)
    E = torch.zeros(nI, device=model.device)

    # initialize xref
    v = torch.bernoulli(torch.sigmoid(v_bias_base).repeat(nI, 1))
    v = torch.bernoulli(torch.rand((nI, model.n_visible), device=model.device))
    h = torch.bernoulli(torch.rand((nI, model.n_hidden), device=model.device))
    E = model.energy(v, h).double().to(model.device)
    print(E.shape)

    for idb in range(1, nb + 1):
        for _ in range(model.k):
            _, h = model.v_to_h(v, blist[idb])
            _, v = model.h_to_v(h, blist[idb])
        E += model.energy(v, h)

    d = 0
    db = 1.0 / nb
    d = -db * E
    d = d.double()
    d0 = torch.mean(d)

    AIS = torch.log(torch.mean(torch.exp(d - d0).double())) + d0
    # AIS = torch.log(torch.mean(torch.exp(Δ)))
    return AIS


def LogLike(model, v, logZ):
    # _, h = model.v_to_h(v)
    F = model.free_energy(v).mean()
    print(F)
    return F - logZ



import torch
import glob
import pickle

from .rbm_pytorch import RBM


# Class for continuous RBM, used for CelebA
class continuous_RBM(RBM):

    def __init__(
        self,
        model_name,
        n_vis=784,
        n_hin=50,
        k=1,
        lr=0.01,
        max_epochs=200000,
        minMax_W=(-100, 100),
        energy_type='hopfield',
        optimizer='Adam',
        regularization=False,
        l1_factor=1e-7,
        l2_factor=1e-7,
        ground_v=1,
        ground_h=1,
        batch_size=64,
        train_algo='CD',
        centering=False,
        average_data=None,
        sampling_model_beta=1,
        mytype=torch.double,
    ):
        # Initialize base class parameters
        super(continuous_RBM, self).__init__(
            model_name,
            n_vis,
            n_vis,
            k,
            lr,
            max_epochs,
            minMax_W,
            energy_type,
            optimizer,
            regularization,
            l1_factor,
            l2_factor,
            ground_v,
            ground_h,
            batch_size,
            train_algo,
            centering,
            average_data,
            sampling_model_beta,
            mytype,
        )
        # Initialize new parameters
        self.logsigma_sq = torch.zeros((n_vis, ),
                                       dtype=self.mytype,
                                       device=self.device)

    def pretrain(self, pretrained_model):
        # Check if you have model load points
        filename_list = glob.glob(
            'model_states/{}_t*.pkl'.format(pretrained_model))
        if len(filename_list) > 0:
            all_loadpoints = sorted([
                int(x.split('_t')[-1].split('.pkl')[0]) for x in filename_list
            ])
            last_epoch = all_loadpoints[-1]
            print('** Using as pretraining model {} at epoch {}'.format(
                pretrained_model, last_epoch))
            with open(
                    'model_states/{}_t{}.pkl'.format(pretrained_model,
                                                     last_epoch),
                    'rb') as file:
                temp_model = pickle.load(file)
                # *** Import pretrained parameters
                self.W = temp_model.W.to(self.mytype)
                self.h_bias = temp_model.h_bias.to(self.mytype)
                self.v_bias = temp_model.v_bias.to(self.mytype)
                self.logsigma_sq = temp_model.logsigma_sq.to(self.mytype)
        else:
            print('** No load points for {}'.format(pretrained_model))

    def v_to_h(self, v, beta=None):
        if beta is None:
            beta = self.model_beta
        return self.GaussBernoulli_v_to_h(v, beta)

    def h_to_v(self, h, beta=None):
        if beta is None:
            beta = self.model_beta
        return self.GaussBernoulli_h_to_v(h, beta)

    def GaussBernoulli_v_to_h(self, v, beta):
        if self.energy_type == 'hopfield':
            p_h = self._prob_h_given_v(v / self.sigma_sq, beta)
        else:
            p_h = self._prob_h_given_v(v, beta)
        sample_h = torch.bernoulli(p_h)
        return p_h, sample_h

    def GaussBernoulli_h_to_v(self, h, beta):
        if self.energy_type == 'hopfield':
            mean = self.delta_ev(h)
            sample_v = mean + torch.randn_like(mean) * self.sigma
        else:
            mean = self.delta_ev(h) * self.sigma_sq
            sample_v = mean + torch.randn_like(mean) * self.sigma

        return None, sample_v

    def tstep_keepup(self):
        if self.energy_type != 'hopfield':
            self.clip_weights()
            self.clip_bias()
            self._update_circuit_variables()
        else:
            self.W_t = self.W.t()
            self.sigma_sq = torch.exp(self.logsigma_sq)
            self.sigma = torch.sqrt(self.sigma_sq)

    def SGD_update(
        self,
        partial_gradient_data,
        partial_gradient_model,
    ):
        # Get the correct gradients
        dEdW_data = partial_gradient_data[0]
        dEdW_model = partial_gradient_model[0]
        dEdv_bias_data = partial_gradient_data[1]
        dEdv_bias_model = partial_gradient_model[1]
        dEdh_bias_data = partial_gradient_data[2]
        dEdh_bias_model = partial_gradient_model[2]
        # Gradients
        dW = -dEdW_data + dEdW_model
        dv = -dEdv_bias_data + dEdv_bias_model
        dh = -dEdh_bias_data + dEdh_bias_model
        if self.energy_type == 'hopfield':
            dEds_data = partial_gradient_data[3]
            dEds_model = partial_gradient_model[3]
            ds = -dEds_data + dEds_model
        # Update parameters in-place
        # and clip
        if self.energy_type == 'hopfield':
            gnorm = torch.norm(dW) + torch.norm(dv) + torch.norm(
                dh) + torch.norm(ds)
        else:
            gnorm = torch.norm(dW) + torch.norm(dv) + torch.norm(dh)
        myclip = (self.lr * 10.0) / gnorm if gnorm > 10 else self.lr
        self.W.add_(dW * myclip)
        self.v_bias.add_(dv * myclip)
        self.h_bias.add_(dh * myclip)
        if self.energy_type == 'hopfield':
            self.logsigma_sq.add_(ds * myclip)

    def generate(self,
                 n_samples,
                 k,
                 h_binarized=True,
                 from_visible=False,
                 beta=None):
        if beta is None:
            beta = self.model_beta
        self.tstep_keepup()
        if from_visible:
            # generate from continuous random visible
            v = torch.randn(size=(n_samples, self.n_visible),
                            device=self.device,
                            dtype=self.mytype)
        else:
            if h_binarized:
                h = torch.randint(
                    high=2,
                    size=(n_samples, self.n_hidden),
                    device=self.device,
                    dtype=self.mytype,
                )
            else:
                h = torch.rand(n_samples,
                               self.n_hidden,
                               device=self.device,
                               dtype=self.mytype)
            _, v = self.h_to_v(h)
        v_model = self.forward(v, k, beta)
        return v_model.detach().cpu().numpy()

    def _update_circuit_variables(self):
        self._k_pos = self.k_pos()
        self._k_neg = self.k_neg()
        self._v_bias_pos = self.v_bias_pos()
        self._v_bias_neg = self.v_bias_neg()
        self._h_bias_pos = self.h_bias_pos()
        self._h_bias_neg = self.h_bias_neg()
        self._k_diff = self._k_pos - self._k_neg
        self._k_diff_t = self._k_diff.t()
        self._v_bias_pos_diff = self._v_bias_pos * (0.5 - self.ground_v)
        self._v_bias_neg_diff = self._v_bias_neg * (0.5 - self.ground_v)
        self._h_bias_pos_diff = self._h_bias_pos * (0.5 - self.ground_h)
        self._h_bias_neg_diff = self._h_bias_neg * (0.5 - self.ground_h)
        self._vb_posneg = -self._v_bias_pos_diff + self._v_bias_neg_diff
        self._hb_posneg = -self._h_bias_pos_diff + self._h_bias_neg_diff
        # for the gaussian units
        self.sigma = torch.pow(
            self._k_pos.sum(0) + self._k_neg.sum(0) + self._v_bias_pos, -0.5)
        self.sigma_sq = self.sigma * self.sigma

    def derivatives_hopfield(self, v, h):
        # h has shape (N, n_h) and v has shape (N, n_v), we want result to have shape (N, n_h, n_v)
        v_diff = v - self.v_bias
        v_over_sigma = v / self.sigma_sq

        if self.centering:
            dEdW = -torch.einsum('ij,ik->ijk', h - self.oh,
                                 v - self.ov) / self.sigma_sq
        else:
            dEdW = -torch.einsum('ij,ik->ijk', h, v_over_sigma)
        dEdv_bias = -(v_diff) / self.sigma_sq
        dEdh_bias = -h
        # dEds = -( (v-self.v_bias)**2/self.sigma_sq.unsqueeze(0)*0.5 - torch.mm(h,self.W)*v/self.sigma_sq.unsqueeze(0))
        dEds = -0.5 * (v_diff**2) / self.sigma_sq + torch.mm(
            h, self.W) * v_over_sigma
        return (dEdW, dEdv_bias, dEdh_bias, dEds)

    def derivatives_linear_circuit_4(self, v, h):
        v_square = torch.square(v).unsqueeze(1)
        vneg_square = torch.square(1 - v).unsqueeze(1)

        h_square = torch.square(h).unsqueeze(2)
        hneg_square = torch.square(1 - h).unsqueeze(2)

        dEdk_pos = (-torch.einsum('ij,ik->ijk', h, v) + 0.5 * v_square + 0.5 *
                    h_square) + (-torch.einsum('ij,ik->ijk', 1 - h, 1 - v) +
                                 0.5 * vneg_square + 0.5 * hneg_square)
        dEdk_neg = (-torch.einsum('ij,ik->ijk', h, 1 - v) + 0.5 * vneg_square +
                    0.5 * h_square) + (-torch.einsum('ij,ik->ijk', 1 - h, v) +
                                       0.5 * v_square + 0.5 * hneg_square)

        dkdw_pos = self.dk_pos().unsqueeze(0)
        dkdw_neg = self.dk_neg().unsqueeze(0)

        dEdW = dEdk_pos * dkdw_pos + dEdk_neg * dkdw_neg

        dEdv_bias_pos = -self.ground_v * v + 0.5 * torch.square(v)
        dEdv_bias_neg = -self.ground_v * (1 - v) + 0.5 * torch.square(1 - v)

        dv_biasdv_pos = self.dv_bias_pos().unsqueeze(0)
        dv_biasdv_neg = self.dv_bias_neg().unsqueeze(0)

        dEdv_bias = dEdv_bias_pos * dv_biasdv_pos + dEdv_bias_neg * dv_biasdv_neg
        dEdh_bias_pos = -self.ground_h * h + 0.5 * torch.square(h)
        dEdh_bias_neg = -self.ground_h * (1 - h) + 0.5 * torch.square(1 - h)

        dh_biasdh_pos = self.dh_bias_pos().unsqueeze(0)
        dh_biasdh_neg = self.dh_bias_neg().unsqueeze(0)
        dEdh_bias = dEdh_bias_pos * dh_biasdh_pos + dEdh_bias_neg * dh_biasdh_neg

        return (dEdW, dEdv_bias, dEdh_bias, None)



import numpy as np
import sys
import matplotlib.pyplot as plt
import pickle
import os
import glob
import torch.fft as fft
import torch
import gzip
import io
from scipy.optimize import fsolve


def load_model(name, delete_previous=False):
    # Check if you have model load points
    filename_list = glob.glob('model_states/{}_t*.pkl'.format(name))
    if len(filename_list) > 0:
        all_loadpoints = sorted(
            [int(x.split('_t')[-1].split('.pkl')[0]) for x in filename_list])
        last_epoch = all_loadpoints[-1]
        print('** Model {} trained up to epoch {}, so I load it'.format(
            name, last_epoch))
        with open('model_states/{}_t{}.pkl'.format(name, last_epoch),
                  'rb') as file:
            model = pickle.load(file)
        if delete_previous:
            # Remove all the previous loadpoints
            for x in all_loadpoints[:-1]:
                os.remove('model_states/{}_t{}.pkl'.format(name, x))
        return True, model
    else:
        print('** No load points for {}'.format(name))
        return False, []


def show_and_save(file_name,
                  img,
                  cmap='gray',
                  vmin=None,
                  vmax=None,
                  save=False,
                  savename=''):
    plt.title(file_name)
    plt.imshow(img, cmap=cmap, vmin=vmin, vmax=vmax)
    if save:
        plt.savefig(savename)
        plt.close()
    else:
        plt.show()


def make_grid(array, nrow=8, padding=2):
    N = array.shape[0]
    H = array.shape[1]
    W = array.shape[2]
    grid_h = int(np.ceil(N / float(nrow)))
    grid_w = nrow
    grid = np.zeros(
        [grid_h * (H + padding) + padding, grid_w * (W + padding) + padding])
    k = 0
    for y in range(grid_h):
        for x in range(grid_w):
            if k < N:
                grid[
                    y * (H + padding):y * (H + padding) + H,
                    x * (W + padding):x * (W + padding) + W,
                ] = array[k]
                k = k + 1
    return grid


def getbasebias(data):
    """Returns the maximum likelihood estimate of the visible bias, given the
    data. If no data is given the RBMs \ bias value is return, but is highly
    recommended to pass the data.

    :return: Base bias.
    :rtype: numpy array [1,  input dim]
    """
    save_mean = torch.clip(data.mean(0), 0.00001, 0.99999)
    return torch.log(save_mean) - torch.log(1.0 - save_mean)


def Covariance_error(centered_data_original, centered_data_model, Nv):
    covariance_matrix_original = torch.matmul(centered_data_original.T,
                                              centered_data_original).mean(
                                                  0)  # / (len(mean_vector)-1)
    covariance_matrix_model = torch.matmul(centered_data_model.T,
                                           centered_data_model).mean(
                                               0)  # / (len(mean_vector)-1)
    return (torch.pow(covariance_matrix_original - covariance_matrix_model,
                      2).sum() * 2 / (Nv * (Nv - 1)))


def Third_moment_error(centered_data_original, centered_data_model, Nv):
    C_ijk_original = (torch.einsum(
        'ni,nj,nk->ijk',
        centered_data_original,
        centered_data_original,
        centered_data_original,
    ) / centered_data_model.shape[0])
    C_ijk_model = (torch.einsum(
        'ni,nj,nk->ijk',
        centered_data_model,
        centered_data_model,
        centered_data_model,
    ) / centered_data_model.shape[0])
    C_ijk = torch.pow(C_ijk_model - C_ijk_original, 2)
    sum_ijk = 0.0
    # m = C_ijk.size(0)
    # for i in range(m):
    #    for j in range(i + 1, m):
    #        for k in range(j + 1, m):
    #            sum_ijk += C_ijk[i, j, k]
    # third_moment_error = sum_ijk*6/(len(mean_vector)*(len(mean_vector)-1)*(len(mean_vector)-2))
    upper_triangular = torch.triu(C_ijk, diagonal=1)
    sum_ijk = upper_triangular.sum(dim=(0, 1, 2))
    return sum_ijk * 6 / (Nv * (Nv - 1) * (Nv - 2))


def PowerSpectrum_MSE(v, v_model):
    # Apply 2D FFT to the signal
    signal_fft_original = fft.fft2(v)
    signal_fft_model = fft.fft2(v_model)
    # Compute the power spectrum
    power_spectrum_original = torch.mean(torch.abs(signal_fft_original)**2, 0)
    power_spectrum_model = torch.mean(torch.abs(signal_fft_model)**2, 0)
    # MSE of the power spectrum
    return torch.mean((torch.log(power_spectrum_original) -
                       torch.log(power_spectrum_model))**2)


def ComputeAATS(v, v_model):
    CONCAT = torch.cat((v, v_model), 1)
    dAB = torch.cdist(CONCAT.t(), CONCAT.t())
    torch.diagonal(dAB).fill_(float('inf'))
    dAB = dAB.cpu().numpy()

    # the next line is use to tranform the matrix into
    #  d_TT d_TF   INTO d_TF- d_TT-  where the minus indicate a reverse order of the columns
    #  d_FT d_FF        d_FT  d_FF
    dAB[:int(dAB.shape[0] / 2), :] = dAB[:int(dAB.shape[0] / 2), ::-1]
    closest = dAB.argmin(axis=1)
    n = int(closest.shape[0] / 2)

    ninv = 1 / n
    AAtruth = (
        (closest[:n] >= n).sum() * ninv
    )  # for a true sample, proba that the closest is in the set of true samples
    AAsyn = (
        (closest[n:] >= n).sum() * ninv
    )  # for a fake sample, proba that the closest is in the set of fake samples

    return AAtruth, AAsyn


# **** Compute FID score
# Generally, if inputting grayscale images into RGB networks, you would just copy
# the grayscale channel three times.
# However, it is unclear how FID behaves in this case, as the original Inception network
# was probably not trained on grayscale images.
# Also, standard FID score uses inception_v3, which is not designed on Mnist,
# but for larger images of realistic scenes.
# So I use for activation an autoencoder that I have trained previously on MNIST.


def Compute_FID(synthetic_images, real_images):
    global inception_model
    try:
        inception_model
    except Exception:
        # Use the following autoencoder for the FID activation
        model_name = 'AE-200000_n2_hs50_SGD_MSE_lr0.01_bs100'
        isloadable, inception_model = load_model(model_name,
                                                 delete_previous=False)
        if not isloadable:
            print('I need {} in order to calculate FID'.format(model_name))
            sys.exit()

    synthetic_activations = (inception_model.encoder(
        synthetic_images.to(torch.double)).detach().cpu().numpy())
    real_activations = (inception_model.encoder(real_images.to(
        torch.double)).detach().cpu().numpy())

    # Compute mean and covariance
    mu_synthetic = np.mean(synthetic_activations)
    mu_real = np.mean(real_activations)

    sigma_synthetic = np.cov(synthetic_activations, rowvar=False)
    sigma_real = np.cov(real_activations, rowvar=False)

    # Compute the FID score
    diff = mu_synthetic - mu_real
    fid_score_flag = False
    if fid_score_flag:
        torch.sqrt(
            torch.trace(sigma_synthetic + sigma_real -
                        2 * torch.sqrt(sigma_synthetic @ sigma_real)))

    # calculate sum squared difference between means
    ssdiff = np.sum((diff)**2.0)
    # calculate sqrt of product between cov
    # Compute eigenvalue decomposition of sigma1.dot(sigma2)
    eigenvalues, eigenvectors = np.linalg.eig(sigma_synthetic.dot(sigma_real))
    # Take the square root of the eigenvalues
    sqrt_eigenvalues = np.sqrt(eigenvalues)
    # Reconstruct the matrix using the square root of eigenvalues and eigenvectors
    covmean = eigenvectors.dot(np.diag(sqrt_eigenvalues)).dot(
        np.linalg.inv(eigenvectors))

    # calculate score
    fid = ssdiff + np.trace(sigma_synthetic + sigma_real - 2.0 * covmean)
    return fid


def Compute_S(v, v_model):
    v = v.detach().cpu().numpy()
    v_model = v_model.detach().cpu().numpy()
    # define a mixed set for crossentropy
    v_cross = v.copy()
    v_cross[:int(0.5 * v_cross.shape[0])] = v_model[:int(0.5 *
                                                         v_cross.shape[0])]

    # Convert the array to bytes
    bytes_io = io.BytesIO()
    np.save(bytes_io, v)
    bytes_src = bytes_io.getvalue()
    np.save(bytes_io, v_cross)
    bytes_cross = bytes_io.getvalue()

    # Compress the bytes using gzip
    compressed_src = gzip.compress(bytes_src)
    compressed_cross = gzip.compress(bytes_cross)

    # Calculate the entropy
    byte_count_src = len(compressed_src)
    byte_count_cross = len(compressed_cross)
    value_counts_src = np.bincount(
        np.frombuffer(compressed_src, dtype=np.uint8))
    value_counts_cross = np.bincount(
        np.frombuffer(compressed_cross, dtype=np.uint8))
    probabilities_src = value_counts_src / byte_count_src
    probabilities_cross = value_counts_cross / byte_count_cross
    entropy_src = -np.sum(probabilities_src * np.log2(probabilities_src))
    entropy_cross = -np.sum(probabilities_cross * np.log2(probabilities_cross))

    return entropy_cross / entropy_src - 1


def generate_S_matrix(shape, target):
    # Generate a random matrix with values between 0 and 1
    random_matrix = np.random.rand(*shape)
    # Adjust the values to achieve the desired average
    adjusted_matrix = random_matrix + (target - np.mean(random_matrix))
    # Clip values to ensure they are between 0 and 1
    adjusted_matrix = np.clip(adjusted_matrix, 0, 1)
    return adjusted_matrix


def generate_synthetic_data(target_entropy, data_size, structured=True):
    # Define the entropy function as a lambda
    def S_lambda(x):
        return -x * np.nan_to_num(np.log2(x)) - (1 - x) * np.nan_to_num(
            np.log2(1 - x))

    # and invert it to get the average pixel that would give you the target entropy
    # (also the target entropy has to be transformed from average number to other thing)
    # I want the target S to come from the average of pixels
    pixel_target = generate_S_matrix((data_size[1], data_size[2]),
                                     target_entropy)
    if structured:
        # To make them into more 'structured' objects, sort the matrix
        flat_indices = np.argsort(pixel_target.flatten())
        # Use the flat indices to reorder the elements in the matrix
        pixel_target = pixel_target.flatten()[flat_indices].reshape(
            pixel_target.shape)
    # Then I look for probabilities that would give me this entropy per pixel
    initial_guess = np.zeros((data_size[1], data_size[2]))
    # Notice that this due to the nature of the fsolve function,
    # the initial_guess breaks the black/white symmetry
    P = fsolve(lambda x: S_lambda(x) - pixel_target.flatten(),
               initial_guess.flatten())  # .reshape(28,28)

    # And finally I use this P to generate the data
    generated_data = ((np.random.rand(data_size[0],
                                      data_size[1] * data_size[2])
                       < P).astype(int)).reshape(data_size)
    S_image, S_pixel = my_entropy(generated_data)
    print(f'\nTarget = {target_entropy}')
    print(f'generated entropy (image) = {S_image.mean()}')
    print(f'generated entropy (pixels) = {S_pixel.mean()}')

    return generated_data


def my_entropy(data):
    X = data.mean(1).mean(1)
    # This is the entropy per image
    S_image = -X * np.nan_to_num(np.log2(X)) - (1 - X) * np.nan_to_num(
        np.log2(1 - X))
    # This is the entropy per pixel
    Y = data.mean(0)
    S_pixel = -Y * np.nan_to_num(np.log2(Y)) - (1 - Y) * np.nan_to_num(
        np.log2(1 - Y))
    return S_image, S_pixel


def binarize_image(image, threshold=128):
    # Binarize the image using a threshold
    binary_image = (image > threshold).astype(int)
    return binary_image



import torch
import numpy as np

from .rbm_pytorch import RBM


# Class for the convolutional RBM model
class convolutional_RBM(RBM):

    def __init__(
        self,
        model_name,
        n_vis=784,
        n_hin=50,
        k=1,
        lr=0.01,
        max_epochs=200000,
        minMax_W=(-100, 100),
        energy_type='hopfield',
        optimizer='Adam',
        regularization=False,
        l1_factor=1e-7,
        l2_factor=1e-7,
        ground_v=1,
        ground_h=1,
        batch_size=64,
        train_algo='CD',
        centering=False,
        average_data=None,
        sampling_model_beta=1,
        mytype=torch.double,
        half_kernel_size=1,
    ):
        # Initialize base class parameters
        super(convolutional_RBM, self).__init__(
            model_name,
            n_vis,
            n_vis,
            k,
            lr,
            max_epochs,
            minMax_W,
            energy_type,
            optimizer,
            regularization,
            l1_factor,
            l2_factor,
            ground_v,
            ground_h,
            batch_size,
            train_algo,
            centering,
            average_data,
            sampling_model_beta,
            mytype,
        )
        # Initialize new parameters
        self.half_kernel_size = half_kernel_size
        # self.W = self.W.to_sparse()

        # Create the adjacency matrix for each input data,
        # assuming that they are stored as row_i + N*column_j
        # * Note that we also assume closed boundaries
        self.nvis_sqrt = int(np.sqrt(self.n_visible))
        self.adjacency_mat = torch.zeros((n_vis, n_vis), device=self.device)
        for i in range(n_vis):
            column = int(np.floor(i / self.nvis_sqrt))
            row = int(i - self.nvis_sqrt * column)
            self.add_neigh(self.adjacency_mat, row, column)
        # self.adjacency_mat = self.adjacency_mat.to_sparse()

    def add_neigh(self, adjacency_mat, row, column):
        my_id = row + column * self.nvis_sqrt
        for i in range(row - self.half_kernel_size,
                       row + self.half_kernel_size + 1):
            for j in range(column - self.half_kernel_size,
                           column + self.half_kernel_size + 1):
                if i >= 0 and i < self.nvis_sqrt and j >= 0 and j < self.nvis_sqrt:
                    neigh_id = i + j * self.nvis_sqrt
                    adjacency_mat[my_id, neigh_id] = 1

    def after_step_keepup(self):
        if isinstance(self, convolutional_RBM):
            self.W = self.adjacency_mat * self.W

        if self.energy_type != 'hopfield':
            self.clip_weights()
            self.clip_bias()
            self._update_circuit_variables()
        else:
            self.W_t = self.W.t()



from setuptools import setup, find_packages

DESCRIPTION = 'Train a RKM'
LONG_DESCRIPTION = 'A Python package to simulate and train a RKM model'

# Setting up
setup(
    name='pyrkm',
    version='0.0.6',
    author='Simone Ciarella',
    author_email='<simoneciarella@gmail.com>',
    url='https://github.com/SCiarella/circuit-rbm',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[
        'numpy',
        'pandas',
        'matplotlib',
        'torch',
        'torchvision',
        'scipy',
    ],
    keywords=['python', 'statistical physics', 'machine learning', 'rbm'],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Education',
        'Programming Language :: Python :: 3',
        'Operating System :: Unix',
    ],
)



import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import time
import pickle


class Autoencoder(nn.Module):

    def __init__(self, model_name, layer_sizes, max_epochs, batch_size, lr,
                 optimizer, loss):
        super(Autoencoder, self).__init__()

        self.name = model_name
        print('*** Initializing {}'.format(self.name))
        self.device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu')
        torch.set_default_dtype(torch.float64)
        print('The model is working on the following device: {}'.format(
            self.device))
        self.max_epochs = max_epochs
        self.lr = lr
        self.epoch = 0
        self.batch_size = batch_size
        self.bottleneck_size = layer_sizes[-1]
        # Epochs at which to store the model
        num_points = 50
        self.t_to_save = sorted(
            list(
                set(
                    np.round(
                        np.logspace(np.log10(1), np.log10(self.max_epochs),
                                    num_points)).astype(int).tolist())))

        if loss == 'MSE':
            self.criterion = nn.MSELoss()
        else:
            print('Loss {} not implemented'.format(loss))

        encoder_layers = []
        for i in range(len(layer_sizes) - 1):
            encoder_layers.append(nn.Linear(layer_sizes[i],
                                            layer_sizes[i + 1]))
            encoder_layers.append(nn.ReLU())

        self.encoder = nn.Sequential(*encoder_layers)

        decoder_layers = []
        for i in range(len(layer_sizes) - 1, 0, -1):
            decoder_layers.append(nn.Linear(layer_sizes[i],
                                            layer_sizes[i - 1]))
            decoder_layers.append(nn.ReLU())

        self.decoder = nn.Sequential(*decoder_layers)

        if optimizer == 'Adam':
            self.optimizer = optim.Adam(self.parameters(), lr=lr)
        elif optimizer == 'SGD':
            self.optimizer = optim.SGD(self.parameters(), lr=lr)
        else:
            print('Optimizer {} not implemented'.format(optimizer))

    def forward(self, x, k=1):
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return decoded

    def train(self, train_data, test_data, print_error=True):
        while self.epoch < self.max_epochs:
            for _, v_data in enumerate(train_data):
                start_time = time.time()
                self.optimizer.zero_grad()
                outputs = self(v_data)
                loss = self.criterion(outputs, v_data)
                loss.backward()
                self.optimizer.step()

                self.epoch += 1

                # Store the model state
                if self.epoch in self.t_to_save:
                    with open(
                            'model_states/{}_t{}.pkl'.format(
                                self.name, self.epoch), 'wb') as file:
                        pickle.dump(self, file)

                if self.epoch % 100 == 0:
                    t = time.time() - start_time
                    if print_error:
                        test_out = self(test_data)
                        test_loss = self.criterion(test_out, test_data)
                        print(
                            'Epoch: %d , Test-loss %.5g , train-loss %.5g , time: %f'
                            % (self.epoch, test_loss, loss, t))
                    else:
                        print('Epoch: %d , train-loss %.5g , time: %f' %
                              (self.epoch, loss, t))

        print('*** Training finished')

    def reconstruct(self, data, k):
        data = torch.Tensor(data).to(self.device).to(torch.double)
        v_model = self(data)
        return data.detach().cpu().numpy(), v_model.detach().cpu().numpy()

    def generate(self, n_samples, k, h_binarized=False, from_visible=False):
        # How does the VAE generates?
        # what is the noise distribution we have to use?
        h = torch.rand(n_samples,
                       self.bottleneck_size).to(self.device).to(torch.double)
        v_model = self.decoder(h)
        return v_model.detach().cpu().numpy()



import torch
import sys

from .rbm_pytorch import RBM


class NReLU(RBM):

    def __init__(
        self,
        model_name,
        n_vis=784,
        n_hin=50,
        k=1,
        lr=0.01,
        max_epochs=200000,
        minMax_W=(-100, 100),
        energy_type='hopfield',
        optimizer='Adam',
        regularization=False,
        l1_factor=1e-7,
        l2_factor=1e-7,
        ground_v=1,
        ground_h=1,
        batch_size=64,
        train_algo='CD',
        centering=False,
        average_data=None,
        sampling_model_beta=1,
        mytype=torch.double,
        half_kernel_size=1,
    ):
        # Initialize base class parameters
        super(NReLU, self).__init__(
            model_name,
            n_vis,
            n_vis,
            k,
            lr,
            max_epochs,
            minMax_W,
            energy_type,
            optimizer,
            regularization,
            l1_factor,
            l2_factor,
            ground_v,
            ground_h,
            batch_size,
            train_algo,
            centering,
            average_data,
            sampling_model_beta,
            mytype,
        )
        # Initialize new parameters
        print('ERROR: NReLU not fully implemented')
        sys.exit()

    def v_to_h(self, v, beta=None):
        if beta is None:
            beta = self.model_beta
        if self.energy_type == 'linear_circuit_5':
            return self.Bernoulli_v_to_h(v, beta)
        else:
            if beta > 1000:
                # I assume we are at T=0
                return self.Deterministic_v_to_h(v, beta)
            else:
                return self.NReLU_v_to_h(v, beta)

    def h_to_v(self, h, beta=None):
        if beta is None:
            beta = self.model_beta
        if self.energy_type == 'linear_circuit_5':
            return self.Bernoulli_h_to_v(h, beta)
        else:
            if beta > 1000:
                # I assume we are at T=0
                return self.Deterministic_h_to_v(h, beta)
            else:
                return self.NReLU_h_to_v(h, beta)

    def NReLU_v_to_h(self, v, beta):
        # out = max(0, x + N(0,sigma(x)))
        x = self.model_beta * self.delta_eh(v)
        activation = x + torch.sqrt(self.my_sigmoid(x)) * torch.randn(
            x.shape, device=self.device)
        sample_h = torch.clip(activation, 0, 1)
        return x, sample_h

    def NReLU_h_to_v(self, h, beta=1):
        x = self.model_beta * self.delta_ev(h)
        # activation = x + x.std()*torch.randn(x.shape, device=self.device)
        activation = x + torch.sqrt(self.my_sigmoid(x)) * torch.randn(
            x.shape, device=self.device)
        sample_v = torch.clip(activation, 0, 1)  # visible is binary
        return x, sample_v

    # The parameters of the 'diode'-sigmoid are hard coded
    def my_sigmoid(self, x, xscale=5, xshift=2.5):
        scaled_x = x * xscale
        shifted_x = scaled_x - xshift
        clipped_x = torch.clip(shifted_x, -10, 10)
        sigmoid_approx = (clipped_x + 10) / 20
        return sigmoid_approx



# This is the __init__.py file for the rkm package

# Import any modules or subpackages here
from .autoencoder import Autoencoder
from .circuit_utils import Circuit
from .rbm_pytorch import RBM
from .convolutional_rbm import convolutional_RBM
from .filtering_rbm import filtering_RBM
from .utils_pytorch import load_model, getbasebias, Covariance_error, Third_moment_error, PowerSpectrum_MSE, ComputeAATS, Compute_FID, Compute_S, show_and_save, make_grid, generate_synthetic_data, my_entropy, binarize_image

__all__ = [
    'Autoencoder', 'Circuit', 'RBM', 'filtering_RBM', 'convolutional_RBM',
    'load_model', 'getbasebias', 'Covariance_error', 'Third_moment_error',
    'PowerSpectrum_MSE', 'ComputeAATS', 'Compute_FID', 'Compute_S',
    'show_and_save', 'make_grid', 'generate_synthetic_data', 'my_entropy',
    'binarize_image'
]
